# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Manage LPAR virtual terminals."""

import re
import select
import six
import socket
import ssl
import struct
import subprocess
import threading
import time

from oslo_concurrency import lockutils as lock
from oslo_log import log as logging

import pypowervm.const as c
from pypowervm import exceptions as pvm_exc
from pypowervm.i18n import _
from pypowervm.wrappers import job
import pypowervm.wrappers.logical_partition as pvm_lpar

LOG = logging.getLogger(__name__)

_SUFFIX_PARM_CLOSE_VTERM = 'CloseVterm'

# Used to track of the mapping between the ports and the Listeners/Repeaters
# that we construct for those and also keeping track of which local port
# is for a given LPAR UUID and want VNC Path String is provided for the LPAR.
#
# These are global variables used below. Since they are defined up here, need
# to use global as a way for modification of the fields to stick.  We do this
# so that we keep track of all of the connections.
_VNC_REMOTE_PORT_TO_LISTENER = {}
_VNC_LOCAL_PORT_TO_REPEATER = {}
_VNC_UUID_TO_LOCAL_PORT = {}
_VNC_PATH_TO_UUID = {}
# For the single remote port case, we will hard-code that to 5901 for now
_REMOTE_PORT = 5901


def close_vterm(adapter, lpar_uuid):
    """Close the vterm associated with an lpar

    :param adapter: The adapter to talk over the API.
    :param lpar_uuid: partition uuid

    """
    if adapter.traits.local_api:
        _close_vterm_local(adapter, lpar_uuid)
    else:
        _close_vterm_non_local(adapter, lpar_uuid)


def _close_vterm_non_local(adapter, lpar_uuid):
    """Job to force the close of the terminal when the API is remote.

    :param adapter: The adapter to talk over the API.
    :param lpar_uuid: partition uuid
    """
    # Close vterm on the lpar
    resp = adapter.read(pvm_lpar.LPAR.schema_type, lpar_uuid,
                        suffix_type=c.SUFFIX_TYPE_DO,
                        suffix_parm=_SUFFIX_PARM_CLOSE_VTERM)
    job_wrapper = job.Job.wrap(resp.entry)

    try:
        job_wrapper.run_job(lpar_uuid)
    except Exception:
        LOG.exception(_('Unable to close vterm.'))
        raise


def _close_vterm_local(adapter, lpar_uuid):
    """Forces the close of the terminal on a local system.

    Will check for a VNC server as well in case it was started via that
    mechanism.

    :param adapter: The adapter to talk over the API.
    :param lpar_uuid: partition uuid
    """
    lpar_id = _get_lpar_id(adapter, lpar_uuid)
    _run_proc(['rmvterm', '--id', lpar_id])

    # Stop the port.
    with lock.lock('powervm_vnc_term'):
        vnc_port = _VNC_UUID_TO_LOCAL_PORT.get(lpar_uuid, 0)
        if vnc_port in _VNC_LOCAL_PORT_TO_REPEATER:
            _VNC_LOCAL_PORT_TO_REPEATER[vnc_port].stop()


def open_localhost_vnc_vterm(adapter, lpar_uuid, force=False):
    """Opens a VNC vTerm to a given LPAR.  Always binds to localhost.

    :param adapter: The adapter to drive the PowerVM API
    :param lpar_uuid: Partition UUID.
    :param force: (Optional, Default: False) If set to true will force the
                  console to be opened as VNC even if it is already opened
                  via some other means.
    :return: The VNC Port that the terminal is running on.
    """
    # This API can only run if local.
    if not adapter.traits.local_api:
        raise pvm_exc.ConsoleNotLocal()

    lpar_id = _get_lpar_id(adapter, lpar_uuid)

    def _run_mkvterm_cmd(lpar_uuid, force):
        cmd = ['mkvterm', '--id', str(lpar_id), '--vnc', '--local']
        ret_code, std_out, std_err = _run_proc(cmd)

        # If the vterm was already started, the mkvterm command will always
        # return an error message with a return code of 3.  However, there
        # are 2 scenarios here, one where it was started with the VNC option
        # previously, which we will get a valid port number back (which is
        # the good path scenario), and one where it was started out-of-band
        # where we will get no port.  If it is the out-of-band scenario and
        # they asked us to force the connection, then we will attempt to
        # terminate the old vterm session so we can start up one with VNC.
        if force and ret_code == 3 and not _parse_vnc_port(std_out):
            LOG.warning(_("Invalid output on vterm open.  Trying to reset the "
                          "vterm.  Error was %s"), std_err)
            close_vterm(adapter, lpar_uuid)
            ret_code, std_out, std_err = _run_proc(cmd)

        # The only error message that is fine is a return code of 3 that a
        # session is already started, where we got back the port back meaning
        # that it was started as VNC.  Else, raise up the error message.
        if ret_code != 0 and not (ret_code == 3 and _parse_vnc_port(std_out)):
            raise pvm_exc.VNCBasedTerminalFailedToOpen(err=std_err)

        # Parse the VNC Port out of the stdout returned from mkvterm
        return _parse_vnc_port(std_out)

    return _run_mkvterm_cmd(lpar_uuid, force)


def open_remotable_vnc_vterm(
        adapter, lpar_uuid, local_ip, remote_ips=None, vnc_path=None,
        use_x509_auth=False, ca_certs=None, server_cert=None, server_key=None,
        force=False):
    """Opens a VNC vTerm to a given LPAR.  Wraps in some validation.

    Must run on the management partition.

    :param adapter: The adapter to drive the PowerVM API
    :param lpar_uuid: Partition UUID.
    :param local_ip: The IP Address to bind the VNC server to.  This would be
                     the IP of the management network on the system.
    :param remote_ips: (Optional, Default: None) A binding to only accept
                       clients that are from a specific list of IP addresses
                       through. Default is None, and therefore will allow any
                       remote IP to connect.
    :param vnc_path: (Optional, Default: None) If provided, the vnc client must
                     pass in this path (in HTTP format) to connect to the
                     VNC server.

                     The path is in HTTP format.  So if the vnc_path is 'Test'
                     the first packet request into the VNC must be:
                     "CONNECT Test HTTP/1.1\r\n\r\n"

                     If the client passes in an invalid request, a 400 Bad
                     Request will be returned.  If the client sends in the
                     correct path a 200 OK will be returned.

                     If no vnc_path is specified, then no path is expected
                     to be passed in by the VNC client and it will listen
                     on the same remote port as local port.  If the path is
                     specified then it will listen on the on a single remote
                     port of 5901 and determine the LPAR based on this path.
    :param use_x509_auth: (Optional, Default: False) If enabled, uses X509
                          Authentication for the VNC sessions started for VMs.
    :param ca_certs: (Optional, Default: None) Path to CA certificate to
                     use for verifying VNC X509 Authentication.  Only used
                     if use_x509_auth is set to True.
    :param server_cert: (Optional, Default: None) Path to Server certificate
                        to use for verifying VNC X509 Authentication.  Only
                        used if use_x509_auth is set to True.
    :param server_key: (Optional, Default: None) Path to Server private key
                       to use for verifying VNC X509 Authentication.  Only
                       used if use_x509_auth is set to True.
    :param force: (Optional, Default: False) If set to true will force the
                  console to be opened as VNC even if it is already opened
                  via some other means.
    :return: The VNC Port that the terminal is running on.
    """
    # This API can only run if local.
    if not adapter.traits.local_api:
        raise pvm_exc.ConsoleNotLocal()

    # Open the VNC Port.  If already open, it will just return the same port,
    # so no harm re-opening.  The stdout will just print out the existing port.
    local_port = open_localhost_vnc_vterm(adapter, lpar_uuid, force=force)
    # If a VNC path is provided then we have a way to map an incoming
    # connection to a given LPAR and will use the single 5901 port, otherwise
    # we need to listen for remote connections on the same port as the local
    # one so we know which VNC session to forward the connection's data to
    remote_port = _REMOTE_PORT if vnc_path is not None else local_port
    _VNC_UUID_TO_LOCAL_PORT[lpar_uuid] = local_port

    # We will use a flag to the Socket Listener to tell it whether the
    # user provided us a VNC Path we should use to look up the UUID from
    if vnc_path is not None:
        verify_vnc_path = True
        _VNC_PATH_TO_UUID[vnc_path] = lpar_uuid
    else:
        verify_vnc_path = False

    # See if we have a VNC repeater already...if so, nothing to do.  If not,
    # start it up.
    with lock.lock('powervm_vnc_term'):
        if remote_port not in _VNC_REMOTE_PORT_TO_LISTENER:
            listener = _VNCSocketListener(
                adapter, remote_port, local_ip, verify_vnc_path,
                remote_ips=remote_ips)
            # If we are doing x509 Authentication, then setup the certificates
            if use_x509_auth:
                listener.set_x509_certificates(
                    ca_certs, server_cert, server_key)
            _VNC_REMOTE_PORT_TO_LISTENER[remote_port] = listener

            listener.start()

    return remote_port


def _run_proc(cmd):
    """Simple wrapper to run a process.

    Will return the return code along with the stdout and stderr.  It is the
    decision of the caller if it wishes to honor or ignore the return code.

    :return: The return code, stdout and stderr from the command.
    """
    process = subprocess.Popen(cmd, shell=False, stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                               close_fds=True, env=None)
    process.wait()
    stdout, stderr = process.communicate()
    return process.returncode, stdout, stderr


def _get_lpar_id(adapter, lpar_uuid):
    lpar_resp = adapter.read(pvm_lpar.LPAR.schema_type, root_id=lpar_uuid,
                             suffix_type='quick', suffix_parm='PartitionID')
    return lpar_resp.body


def _parse_vnc_port(std_out):
    """Parse the VNC port number out of the standard output from mkvterm.

    :return:  The port number parsed otherwise None if no valid port
    """
    # The first line of the std_out should be the VNC port
    line = std_out.splitlines()[0] if std_out else None
    return int(line) if line and line.isdigit() else None


class _VNCSocketListener(threading.Thread):
    """Provides a listener bound to a remote-accessible port for VNC access.

    The VNC sessions set up by mkvterm only allow access from the localhost, so
    this listener provides an additional listener on a remote-accessible port
    to all incoming connections for VNC sessions.

    This listener may be setup by the caller in a way so that there is only a
    single remote port for all VNC sessions or that there is one port per VM.
    This listener will accept incoming connections, establish authentication of
    the requester (if x509 authentication is enabled), and will determine what
    LPAR UUID the request is for and establish connections to the local port
    and setup a repeater to forward the data between the two sides.
    """

    def __init__(self, adapter, remote_port, local_ip, verify_vnc_path,
                 remote_ips=None):
        """Creates the listener bound to a remote-accessible port.

        :param adapter: The pypowervm adapter
        :param remote_port: The port to bind to for remote connections.
        :param local_ip: The IP address to bind the VNC server to. This would
                         be the IP of the management network on the system.
        :param verify_vnc_path: Boolean to determine whether we verify the
                                vnc_path.
        :param remote_ips: (Optional, Default: None) A binding to only accept
                           clients that are from a specific list of IP
                           addresses through. Default is None, and therefore
                           will allow any remote IP to connect.
        """
        super(_VNCSocketListener, self).__init__()

        self.adapter = adapter
        self.remote_port = remote_port
        self.local_ip = local_ip
        self.verify_vnc_path = verify_vnc_path
        self.remote_ips = remote_ips
        self.x509_certs = None

        self.alive = True
        self.vnc_killer = None

    def set_x509_certificates(self, ca_certs=None,
                              server_cert=None, server_key=None):
        """Set the x509 Certificates to use for TLS authentication.

        :param ca_certs: (Optional, Default: None) Path to CA certificate to
                         use for verifying VNC X509 Authentication.
        :param server_cert: (Optional, Default: None) Path to Server cert
                            to use for verifying VNC X509 Authentication.
        :param server_key: (Optional, Default: None) Path to Server private key
                           to use for verifying VNC X509 Authentication.
        """
        self.x509_certs = dict(
            ca_certs=ca_certs, server_cert=server_cert, server_key=server_key)

    def stop(self):
        """Stops the listener from running."""
        # This will stop listening for all clients
        self.alive = False

        # Remove ourselves from the VNC listeners.
        if self.remote_port in _VNC_REMOTE_PORT_TO_LISTENER:
            del _VNC_REMOTE_PORT_TO_LISTENER[self.remote_port]

    def run(self):
        """Used by the thread to run the listener."""
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((self.local_ip, self.remote_port))
        LOG.info(_("VNCSocket Listener Listening on ip=%(ip)s port=%(port)s") %
                 {'ip': self.local_ip, 'port': self.remote_port})
        server.listen(10)

        while self.alive:
            # Listen on the server socket for incoming connections
            s_inputs = select.select([server], [], [], 1)[0]
            for s_input in s_inputs:
                # Establish a new client connection & repeater between the two
                self._new_client(s_input)
        server.close()

    def _new_client(self, server):
        """Listens for a new client.

        :param server: The server socket.
        """
        # This is the socket FROM the client side.  client_addr is a tuple
        # of format ('1.2.3.4', '5678') - ip and port.
        client_socket, client_addr = server.accept()
        LOG.debug("New Client socket accepted client_addr=%s" % client_addr[0])

        # If only select IPs are allowed through, validate
        if (self.remote_ips is not None and
                client_addr[0] not in self.remote_ips):
            # Close the connection, exit.
            client_socket.close()
            return

        # If they gave use a VNC Path to look for in the connection string
        # then we will do that now otherwise just skip over the header info
        if self.verify_vnc_path:
            # Check to ensure that there is output waiting.
            c_input = select.select([client_socket], [], [], 1)[0]

            # If no input, then just assume a close.  We waited a second.
            if not c_input:
                # Assume HTTP 1.1.  All clients should support.  We have no
                # input, so we don't know what protocol they would like.
                client_socket.sendall("HTTP/1.1 400 Bad Request\r\n\r\n")
                client_socket.close()
                return

            # We know we had data waiting.  Receive (at max) the vnc_path
            # string.  All data after this validation string is the
            # actual VNC data.
            lpar_uuid, http_code = self._check_http_connect(client_socket)
            if lpar_uuid:
                # Send back the success message.
                client_socket.sendall("HTTP/%s 200 OK\r\n\r\n" % http_code)
            else:
                # Was not a success, exit.
                client_socket.sendall("HTTP/%s 400 Bad Request\r\n\r\n" %
                                      http_code)
                client_socket.close()
                return
        # If we had no VNC Path to match against, then the local port is
        # going to be the same as the remote port and we need to figure
        # out what the LPAR UUID is for that given local port VNC session
        else:
            lpar_uuid = (k for k, v in _VNC_UUID_TO_LOCAL_PORT.items()
                         if v == self.remote_port).next()

        # Setup the forwarding socket to the local LinuxVNC session
        self._setup_forwarding_socket(lpar_uuid, client_socket)

    def _setup_forwarding_socket(self, lpar_uuid, client_socket):
        """Setup the forwarding socket to the local LinuxVNC session.

        :param lpar_uuid: The UUID of the lpar for which we are forwarding.
        :param client_socket:  The client-side socket to receive data from.
        """
        local_port = _VNC_UUID_TO_LOCAL_PORT.get(lpar_uuid)

        # If for some reason no mapping to a local port, then give up
        if local_port is None:
            client_socket.close()
        # Get the forwarder.  This will be the socket we read FROM the
        # localhost.  When this receives data, it will be sent to the client
        # socket.
        fwd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        fwd.connect(('127.0.0.1', local_port))

        # If we were told to enable VeNCrypt using X509 Authentication, do so
        if self.x509_certs is not None:
            ssl_socket = self._enable_x509_authentication(client_socket, fwd)
            # If there was an error enabling SSL, then close the sockets
            if ssl_socket is None:
                client_socket.close()
                fwd.close()
                return
            client_socket = ssl_socket

        # See if we need to start up a new repeater for the given local port
        if local_port not in _VNC_LOCAL_PORT_TO_REPEATER:
            _VNC_LOCAL_PORT_TO_REPEATER[local_port] = _VNCRepeaterServer(
                self.adapter, lpar_uuid, local_port, client_socket, fwd)
            _VNC_LOCAL_PORT_TO_REPEATER[local_port].start()
        else:
            repeater = _VNC_LOCAL_PORT_TO_REPEATER[local_port]
            repeater.add_socket_connection_pair(client_socket, fwd)

    def _enable_x509_authentication(self, client_socket, server_socket):
        """Enables and Handshakes VeNCrypt using X509 Authentication.

        :param client_socket:  The client-side socket to receive data from.
        :param server_socket:  The server-side socket to forward data to.
        :return ssl_socket:  A client-side socket wrappered for SSL or None
                             if there is an error.
        """
        try:
            # First perform the RFB Version negotiation between client/server
            self._version_negotiation(client_socket, server_socket)
            # Next perform the Security Authentication Type Negotiation
            if not self._auth_type_negotiation(client_socket):
                return None
            # Next perform the Security Authentication SubType Negotiation
            if not self._auth_subtype_negotiation(client_socket):
                return None
            # Now that the VeNCrypt handshake is done, do the SSL wrapper
            ca_certs = self.x509_certs.get('ca_certs')
            server_key = self.x509_certs.get('server_key')
            server_cert = self.x509_certs.get('server_cert')
            return ssl.wrap_socket(
                client_socket, server_side=True, ca_certs=ca_certs,
                certfile=server_cert, keyfile=server_key,
                ssl_version=ssl.PROTOCOL_TLSv1_2, cert_reqs=ssl.CERT_REQUIRED)
        # If we got an error, log and handle to not take down the thread
        except Exception as exc:
            LOG.warning(_("Error negotiating SSL for VNC Repeater: %s") % exc)
            LOG.exception(exc)
            return None

    def _version_negotiation(self, client_socket, server_socket):
        """Performs the RFB Version negotiation between client/server.

        :param client_socket:  The client-side socket to receive data from.
        :param server_socket:  The server-side socket to forward data to.
        """
        # Do a pass-thru of the RFB Version negotiation up-front
        # The length of the version is 12, such as 'RFB 003.007\n'
        client_socket.sendall(self._socket_receive(server_socket, 12))
        server_socket.sendall(self._socket_receive(client_socket, 12))
        # Since we are doing our own additional authentication
        # just tell the server we are doing No Authentication (1) to it
        auth_size = self._socket_receive(server_socket, 1)
        self._socket_receive(server_socket, six.byte2int(auth_size))
        server_socket.sendall(six.int2byte(1))

    def _auth_type_negotiation(self, client_socket):
        """Performs the VeNCrypt Authentication Type Negotiation.

        :param client_socket:  The client-side socket to receive data from.
        :return success:  Boolean whether the handshake was successful.
        """
        # Do the VeNCrypt handshake next before establishing SSL
        # Say we only support VeNCrypt (19) authentication version 0.2
        client_socket.sendall(six.int2byte(1))
        client_socket.sendall(six.int2byte(19))
        client_socket.sendall("\x00\x02")
        authtype = self._socket_receive(client_socket, 1)
        # Make sure the Client supports the VeNCrypt (19) authentication
        if len(authtype) < 1 or six.byte2int(authtype) != 19:
            # Send a 1 telling the client the type wasn't accepted
            client_socket.sendall(six.int2byte(1))
            return False
        vers = self._socket_receive(client_socket, 2)
        # Make sure the Client supports at least version 0.2 of it
        if ((len(vers) < 2 or six.byte2int(vers) != 0
             or six.byte2int(vers[1:]) < 2)):
            # Send a 1 telling the client the type wasn't accepted
            client_socket.sendall(six.int2byte(1))
            return False
        # Tell the Client we have accepted the authentication type
        # In this particular case 0 means the type was accepted
        client_socket.sendall(six.int2byte(0))
        return True

    def _auth_subtype_negotiation(self, client_socket):
        """Performs the x509None Authentication Sub-Type Negotiation.

        :param client_socket:  The client-side socket to receive data from.
        :return success:  Boolean whether the handshake was successful.
        """
        # Tell the client the authentication sub-type is x509None (260)
        client_socket.sendall(six.int2byte(1))
        client_socket.sendall(struct.pack('!I', 260))
        subtyp_raw = self._socket_receive(client_socket, 4)
        # Make sure that the client also supports sub-type x509None (260)
        if 260 not in struct.unpack('!I', subtyp_raw):
            # Send a 0 telling the client the sub-type wasn't accepted
            client_socket.sendall(six.int2byte(0))
            return False
        # Tell the Client we have accepted the authentication handshake
        # In this particular case 1 means the sub-type was accepted
        client_socket.sendall(six.int2byte(1))
        return True

    def _socket_receive(self, asocket, bufsize):
        """Helper method to add a timeout on each receive call.

        This method will raise a timeout exception if it takes > 30 seconds.

        :param asocket: The socket to do the receive on.
        :param bufsize: The number of bytes to receive.
        :return data: The data returned from the socket receive.
        """
        # Add a 30 second timeout around the receive so that we don't
        # block forever if for some reason it never received the packet
        if not select.select([asocket], [], [], 30)[0]:
            raise socket.timeout('30 second timeout on handshake receive')
        return asocket.recv(bufsize)

    def _check_http_connect(self, client_socket):
        """Parse the HTTP connect string to find the LPAR UUID.

        :param client_socket: The client socket sending the data.
        :returns lpar_uuid: The LPAR UUID parsed from the connect string.
        :returns http_code: The HTTP Connection code used for the client
                            connection.
        """
        # Get the expected header.
        # We don't know how large the identifier will be, so use 500 as max.
        # If the identifier is less than 500, it will not return as many bytes.
        header_len = len("CONNECT %s HTTP/1.1\r\n\r\n" % ('x' * 500))
        value = client_socket.recv(header_len)

        # Find the HTTP Code (if you can...)
        pat = r'^CONNECT\s+(\S+)\s+HTTP/(.*)\r\n\r\n$'
        res = re.match(pat, value)
        vnc_path = res.groups()[0] if res else None
        http_code = res.groups()[1] if res else '1.1'
        return _VNC_PATH_TO_UUID.get(vnc_path), http_code


class _VNCRepeaterServer(threading.Thread):
    """Repeats a VNC connection from localhost to a given client.

    This class is separated out from the Socket Listener so that there can
    be one thread doing the actual repeating/forwarded of the data for the
    VNC sessions for a single LPAR.  Otherwise if there are sessions to a lot
    of LPAR's with sessions, one overall thread might get overloaded.

    This class will be provided a pair of peer socket connections and will
    listen for data from each of them and forward to the other until the
    connection on one side goes down in which it will close the connection
    to the other side.

    Also, if no connections are open for a given local port VNC session,
    after a 5 minute window it will run rmvterm to close the terminal console
    to clean up sessions that are no longer being used.
    """

    def __init__(self, adapter, lpar_uuid, local_port, client_socket=None,
                 local_socket=None):
        """Creates the repeater.

        :param adapter: The pypowervm adapter
        :param lpar_uuid: Partition UUID.
        :param local_port: The local port bound to by the VNC session.
        :param client_socket: (Optional, Default: None) The socket descriptor
                              of the incoming client connection.
        :param local_socket: (Optional, Default: None) The socket descriptor of
                             the VNC session connection forwarding data to.
        """
        super(_VNCRepeaterServer, self).__init__()

        self.peers = dict()
        self.adapter = adapter
        self.lpar_uuid = lpar_uuid
        self.local_port = local_port
        self.alive = True
        self.vnc_killer = None

        # Add the connection passed into us to the forwarding list
        if client_socket is not None and local_socket is not None:
            self.add_socket_connection_pair(client_socket, local_socket)

    def stop(self):
        """Stops the repeater from running."""
        # This will stop listening for all clients
        self.alive = False

        # Remove ourselves from the VNC listeners.
        if self.local_port in _VNC_LOCAL_PORT_TO_REPEATER:
            del _VNC_LOCAL_PORT_TO_REPEATER[self.local_port]

    def run(self):
        """Used by the thread to run the repeater."""
        while self.alive:
            # Do a select to wait for data on each of the socket connections
            input_list = list(self.peers)
            s_inputs = select.select(input_list, [], [], 1)[0]

            for s_input in s_inputs:
                # At this point, we need to read the data.  We know that data
                # is ready.  However, if that data that is ready is length
                # 0, then we know that we're ready to close this.
                data = s_input.recv(4096)
                if len(data) == 0:
                    self._close_client(s_input)

                    # Note that we have to break here.  We do that because the
                    # peer dictionary has changed with the close.  So the list
                    # to iterate over should be re-evaluated.
                    # The remaining inputs will just be picked up on the next
                    # pass, so nothing to worry about.
                    break

                # Just process the data.
                self.peers[s_input].send(data)

        # At this point, force a close on all remaining inputs.
        for input_socket in self.peers:
            input_socket.close()

    def add_socket_connection_pair(self, client_socket, local_socket):
        """Adds the pair of socket connections to the list to forward data for.

        :param client_socket: The client-side incoming socket.
        :param local_socket: The local socket for the VNC session.
        """
        self.peers[local_socket] = client_socket
        self.peers[client_socket] = local_socket
        # If for some reason the VNC was being killed, abort it
        if self.vnc_killer is not None:
            self.vnc_killer.abort()
            self.vnc_killer = None

    def _close_client(self, s_input):
        """Closes down a client.

        :param s_input: The socket that has received a close.
        """
        # Close the sockets
        peer = self.peers[s_input]
        peer.close()
        s_input.close()

        # And remove from the peer list, so that we've removed all pointers to
        # them
        del self.peers[peer]
        del self.peers[s_input]

        # If this was the last port, close the local connection
        if len(self.peers) == 0:
            self.vnc_killer = _VNCKiller(self.adapter, self.lpar_uuid)
            self.vnc_killer.start()


class _VNCKiller(threading.Thread):
    """The VNC Killer is a thread that will eventually close the VNC.

    The VNC Repeater could run indefinitely, whether clients are connected to
    it or not.  This class will wait a period of time (5 minutes) and if
    the abort has not been called, will fully close the vterm.

    This is used in orchestration with the VNCRepeaterServer.  The intention
    is, if the user quickly navigates off the VNC, they can come back without
    losing their whole session.  But if they wait up to 5 minutes, then the
    session will be closed out and the memory will be reclaimed.
    """

    def __init__(self, adapter, lpar_uuid):
        super(_VNCKiller, self).__init__()
        self.adapter = adapter
        self.lpar_uuid = lpar_uuid
        self._abort = False

    def abort(self):
        """Call to stop the killer from completing its job."""
        self._abort = True

    def run(self):
        count = 0

        # Wait up to 5 minutes to see if any new negotiations came in
        while count < 300 and not self._abort:
            time.sleep(1)
            if self._abort:
                break
            count += 1

        if not self._abort:
            _close_vterm_local(self.adapter, self.lpar_uuid)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
import mock
import unittest
import uuid

import pypowervm.entities as ent
import pypowervm.tasks.cluster_ssp as cs
import pypowervm.tasks.storage as tsk_st
import pypowervm.tests.tasks.util as tju
from pypowervm.tests.test_utils import test_wrapper_abc as twrap
import pypowervm.util as u
import pypowervm.wrappers.cluster as clust
import pypowervm.wrappers.job as jwrap
import pypowervm.wrappers.storage as stor

CREATE_CLUSTER = 'cluster_create_job_template.txt'


class TestClusterSSP(unittest.TestCase):

    @mock.patch('pypowervm.wrappers.job.Job.delete_job')
    @mock.patch('pypowervm.wrappers.job.Job._monitor_job')
    @mock.patch('pypowervm.wrappers.job.Job.job_status')
    @mock.patch('pypowervm.adapter.Adapter')
    def test_crt_cluster_ssp(self, mock_adp, mock_status, mock_monitor_job,
                             mock_del_job):
        # Load up GET Cluster/do/Create (job template)
        mock_adp.read.return_value = tju.load_file(CREATE_CLUSTER, mock_adp)
        # We'll pretend the job ran and completed successfully
        mock_monitor_job.return_value = False
        mock_status.__get__ = mock.Mock(
            return_value=jwrap.JobStatus.COMPLETED_OK)

        # Mock Job.create_job to check job parameter values
        def create_job(job_el, entry_type, *args, **kwargs):
            self.assertEqual(entry_type, clust.Cluster.schema_type)
            job = jwrap.Job.wrap(ent.Entry({}, job_el, None))
            param_vals = job._get_vals(u.xpath(
                'JobParameters', 'JobParameter', 'ParameterValue'))
            self.assertEqual(
                param_vals[0],
                '<uom:Cluster xmlns:uom="http://www.ibm.com/xmlns/systems/powe'
                'r/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadat'
                'a><uom:Atom/></uom:Metadata><uom:ClusterName>clust_name</uom:'
                'ClusterName><uom:RepositoryDisk schemaVersion="V1_0"><uom:Phy'
                'sicalVolume schemaVersion="V1_0"><uom:Metadata><uom:Atom/></u'
                'om:Metadata><uom:VolumeName>repos_pv_name</uom:VolumeName></u'
                'om:PhysicalVolume></uom:RepositoryDisk><uom:Node schemaVersio'
                'n="V1_0"><uom:Node schemaVersion="V1_0"><uom:Metadata><uom:At'
                'om/></uom:Metadata><uom:HostName>vios1</uom:HostName><uom:Par'
                'titionID>5</uom:PartitionID><uom:MachineTypeModelAndSerialNum'
                'ber schemaVersion="V1_0"><uom:Metadata><uom:Atom/></uom:Metad'
                'ata><uom:MachineType>XXXX</uom:MachineType><uom:Model>YYY</uo'
                'm:Model><uom:SerialNumber>ZZZZZZZ</uom:SerialNumber></uom:Mac'
                'hineTypeModelAndSerialNumber><uom:VirtualIOServer href="https'
                '://a.example.com:12443/rest/api/uom/VirtualIOServer/12345678-'
                '1234-1234-1234-123456789012" rel="related"/></uom:Node></uom:'
                'Node></uom:Cluster>')
            self.assertEqual(
                param_vals[1],
                '<uom:SharedStoragePool xmlns:uom="http://www.ibm.com/xmlns/sy'
                'stems/power/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><u'
                'om:Metadata><uom:Atom/></uom:Metadata><uom:PhysicalVolumes sc'
                'hemaVersion="V1_0"><uom:PhysicalVolume schemaVersion="V1_0"><'
                'uom:Metadata><uom:Atom/></uom:Metadata><uom:VolumeName>hdisk1'
                '</uom:VolumeName></uom:PhysicalVolume><uom:PhysicalVolume sch'
                'emaVersion="V1_0"><uom:Metadata><uom:Atom/></uom:Metadata><uo'
                'm:VolumeName>hdisk2</uom:VolumeName></uom:PhysicalVolume><uom'
                ':PhysicalVolume schemaVersion="V1_0"><uom:Metadata><uom:Atom/'
                '></uom:Metadata><uom:VolumeName>hdisk3</uom:VolumeName></uom:'
                'PhysicalVolume></uom:PhysicalVolumes><uom:StoragePoolName>ssp'
                '_name</uom:StoragePoolName></uom:SharedStoragePool>')
            return mock.MagicMock()
        mock_adp.create_job.side_effect = create_job
        node = clust.Node.bld(
            mock_adp, hostname='vios1', lpar_id=5, mtms='XXXX-YYY*ZZZZZZZ',
            vios_uri='https://a.example.com:12443/rest/api/uom/VirtualIOServe'
            'r/12345678-1234-1234-1234-123456789012')
        repos = stor.PV.bld(mock_adp, name='repos_pv_name')
        data = [stor.PV.bld(mock_adp, name=n) for n in (
            'hdisk1', 'hdisk2', 'hdisk3')]
        cs.crt_cluster_ssp('clust_name', 'ssp_name', repos, node, data)
        # run_job() should run delete_job() at the end
        self.assertEqual(mock_del_job.call_count, 1)


class TestGetOrUploadImageLU(twrap.TestWrapper):
    file = 'lufeed.txt'
    wrapper_class_to_test = stor.LUEnt

    def setUp(self):
        super(TestGetOrUploadImageLU, self).setUp()
        self.tier = mock.Mock(spec=stor.Tier)
        self.mock_luent_srch = self.useFixture(fixtures.MockPatch(
            'pypowervm.wrappers.storage.LUEnt.search')).mock
        self.mock_luent_srch.side_effect = self.luent_search
        self.useFixture(fixtures.MockPatch(
            'uuid.uuid4')).mock.return_value = uuid.UUID('1234abcd-1234-1234-1'
                                                         '234-abcdbcdecdef')
        self.mock_crt_lu = self.useFixture(fixtures.MockPatch(
            'pypowervm.tasks.storage.crt_lu')).mock
        self.mock_upload_lu = self.useFixture(fixtures.MockPatch(
            'pypowervm.tasks.storage.upload_lu')).mock
        self.mock_upload_lu.side_effect = self.upload_lu
        self.mock_sleep = self.useFixture(fixtures.MockPatch(
            'time.sleep')).mock
        self.mock_sleep.side_effect = self.sleep_conflict_finishes
        self.vios_uuid = 'vios_uuid'
        self.mock_stream_func = mock.Mock()
        self.gb_size = 123
        self.b_size = self.gb_size * 1024 * 1024 * 1024
        # The image LU with the "real" content
        luname = 'lu_name'
        self.img_lu = self.bld_lu(luname, self.gb_size)
        # The marker LU used by *this* thread
        mkrname = 'part1234abcd' + luname
        self.mkr_lu = self.bld_lu(mkrname, cs.MKRSZ)
        # Marker LU used by a conflicting thread.  This one will lose the bid.
        confl_luname_lose = 'part5678cdef' + luname
        self.confl_mkr_lu_lose = self.bld_lu(confl_luname_lose, cs.MKRSZ)
        # Marker LU used by a conflicting thread.  This one will win the bid.
        confl_luname_win = 'part0123abcd' + luname
        self.confl_mkr_lu_win = self.bld_lu(confl_luname_win, cs.MKRSZ)
        # Always expect to finish with exactly one more LU than we started with
        self.exp_num_lus = len(self.entries) + 1

    def bld_lu(self, luname, gb_size):
        lu = stor.LUEnt.bld(None, luname, gb_size, typ=cs.IMGTYP)
        lu._udid('udid_' + luname)
        lu.delete = mock.Mock()
        lu.delete.side_effect = lambda: self.entries.remove(lu)
        return lu

    def luent_search(self, adapter, parent=None, lu_type=None):
        """Mock side effect for LUEnt.search, validating arguments.

        :return: self.entries (the LUEnt feed)
        """
        self.assertEqual(self.tier.adapter, adapter)
        self.assertEqual(self.tier, parent)
        self.assertEqual(cs.IMGTYP, lu_type)
        return self.entries

    def setup_crt_lu_mock(self, crt_img_lu_se, conflicting_mkr_lu=None):
        """Set up the mock side effect for crt_lu calls.

        The marker LU side always creates "my" marker LU.  If a
        conflicting_mkr_lu is specified, also creates that marker LU (to
        simulate simultaneous attempts from separate hosts).

        The image LU side behaves as indicated by the crt_img_lu_se parameter.

        :param crt_img_lu_se: Side effect for crt_lu of the image LU.
        :param conflicting_mkr_lu: If specified, the resulting mock pretends
                                   that some other host created the specified
                                   marker LU at the same time we're creating
                                   ours.
        :return: A callable suitable for assigning to
                 self.mock_crt_lu.side_effect.
        """
        def crt_mkr_lu(tier, luname, lu_gb, typ=None):
            self.assertEqual(self.tier, tier)
            self.assertEqual(self.mkr_lu.name, luname)
            self.assertEqual(self.mkr_lu.capacity, lu_gb)
            self.assertEqual(cs.IMGTYP, typ)
            self.entries.append(self.mkr_lu)
            if conflicting_mkr_lu is not None:
                self.entries.append(conflicting_mkr_lu)
            # Second time through, creation of the image LU
            self.mock_crt_lu.side_effect = crt_img_lu_se
            return tier, self.mkr_lu

        # First time through, creation of the marker LU
        self.mock_crt_lu.side_effect = crt_mkr_lu

    def crt_img_lu(self, tier, luname, lu_gb, typ=None):
        """Mock side effect for crt_lu of the image LU."""
        self.assertEqual(self.tier, tier)
        self.assertEqual(self.img_lu.name, luname)
        self.assertEqual(self.img_lu.capacity, lu_gb)
        self.assertEqual(cs.IMGTYP, typ)
        self.entries.append(self.img_lu)
        return tier, self.img_lu

    def upload_lu(self, vios_uuid, new_lu, stream, b_size, upload_type=None):
        self.assertEqual(self.vios_uuid, vios_uuid)
        self.assertEqual(self.img_lu, new_lu)
        self.assertEqual(self.mock_stream_func, stream)
        self.assertEqual(self.b_size, b_size)
        self.assertEqual(tsk_st.UploadType.IO_STREAM_BUILDER, upload_type)

    def sleep_conflict_finishes(self, sec):
        """Pretend the conflicting LU finishes while we sleep."""
        self.assertTrue(cs.SLEEP_U_MIN <= sec <= cs.SLEEP_U_MAX)
        # We may have used either conflict marker LU
        if self.confl_mkr_lu_lose in self.entries:
            self.entries.remove(self.confl_mkr_lu_lose)
        if self.confl_mkr_lu_win in self.entries:
            self.entries.remove(self.confl_mkr_lu_win)
        if self.img_lu not in self.entries:
            self.entries.append(self.img_lu)

    def test_already_exists(self):
        """The image LU is already there."""
        self.entries.append(self.img_lu)

        self.assertEqual(self.img_lu, cs.get_or_upload_image_lu(
            self.tier, self.img_lu.name, self.vios_uuid, self.mock_stream_func,
            self.b_size))

        # We only searched once
        self.assertEqual(1, self.mock_luent_srch.call_count)
        # We didn't create anything
        self.mock_crt_lu.assert_not_called()
        # We didn't upload anything
        self.mock_upload_lu.assert_not_called()
        # We didn't delete anything
        self.mkr_lu.delete.assert_not_called()
        self.img_lu.delete.assert_not_called()
        # We didn't sleep
        self.mock_sleep.assert_not_called()
        # Stream func not invoked
        self.mock_stream_func.assert_not_called()
        # Right number of LUs
        self.assertEqual(self.exp_num_lus, len(self.entries))

    def test_upload_no_conflict(self):
        """Upload a new LU - no conflict."""
        self.setup_crt_lu_mock(self.crt_img_lu)

        self.assertEqual(self.img_lu, cs.get_or_upload_image_lu(
            self.tier, self.img_lu.name, self.vios_uuid, self.mock_stream_func,
            self.b_size))

        # Uploaded content
        self.assertEqual(1, self.mock_upload_lu.call_count)
        # Removed marker LU
        self.mkr_lu.delete.assert_called_once_with()
        # Did not delete image LU
        self.img_lu.delete.assert_not_called()
        # I pulled the feed the first time through, and for _upload_conflict
        self.assertEqual(2, self.mock_luent_srch.call_count)
        # Right number of LUs
        self.assertEqual(self.exp_num_lus, len(self.entries))

    def test_conflict_not_started(self):
        """Another upload is about to start when we get there."""
        # Note that the conflicting process wins, even though its marker LU
        # name would lose to ours - because we don't get around to creating
        # ours.
        self.entries.append(self.confl_mkr_lu_lose)

        self.assertEqual(self.img_lu, cs.get_or_upload_image_lu(
            self.tier, self.img_lu.name, self.vios_uuid, self.mock_stream_func,
            self.b_size))

        # I "waited" for the other guy to complete
        self.assertEqual(1, self.mock_sleep.call_count)
        # I did not create, upload, or remove anything
        self.mock_crt_lu.assert_not_called()
        self.mock_upload_lu.assert_not_called()
        self.mkr_lu.delete.assert_not_called()
        self.img_lu.delete.assert_not_called()
        # I pulled the feed the first time through, and once after the sleep
        self.assertEqual(2, self.mock_luent_srch.call_count)
        # Right number of LUs
        self.assertEqual(self.exp_num_lus, len(self.entries))

    def test_conflict_started(self):
        """Another upload is in progress when we get there."""
        self.entries.append(self.confl_mkr_lu_lose)
        self.entries.append(self.img_lu)

        self.assertEqual(self.img_lu, cs.get_or_upload_image_lu(
            self.tier, self.img_lu.name, self.vios_uuid, self.mock_stream_func,
            self.b_size))

        # I "waited" for the other guy to complete
        self.assertEqual(1, self.mock_sleep.call_count)
        # I did not create, upload, or remove anything
        self.mock_crt_lu.assert_not_called()
        self.mock_upload_lu.assert_not_called()
        self.mkr_lu.delete.assert_not_called()
        self.img_lu.delete.assert_not_called()
        # I searched the first time through, and once after the sleep
        self.assertEqual(2, self.mock_luent_srch.call_count)
        # Right number of LUs
        self.assertEqual(self.exp_num_lus, len(self.entries))

    def test_conflict_I_lose(self):
        """We both bid at the same time; and I lose."""
        self.setup_crt_lu_mock(self.fail,
                               conflicting_mkr_lu=self.confl_mkr_lu_win)

        self.assertEqual(self.img_lu, cs.get_or_upload_image_lu(
            self.tier, self.img_lu.name, self.vios_uuid, self.mock_stream_func,
            self.b_size))

        # I tried creating mine because his wasn't there at the start
        self.assertEqual(1, self.mock_crt_lu.call_count)
        # I "slept", waiting for the other guy to finish
        self.assertEqual(1, self.mock_sleep.call_count)
        # I didn't upload
        self.mock_upload_lu.assert_not_called()
        # I did remove my marker from the SSP
        self.mkr_lu.delete.assert_called_once_with()
        # I didn't remove the image LU (because I didn't create it)
        self.img_lu.delete.assert_not_called()
        # I searched the first time through, once in _upload_conflict, and once
        # after the sleep
        self.assertEqual(3, self.mock_luent_srch.call_count)
        # Right number of LUs
        self.assertEqual(self.exp_num_lus, len(self.entries))

    def test_conflict_I_win(self):
        """We both bid at the same time; and I win."""
        self.setup_crt_lu_mock(self.crt_img_lu,
                               conflicting_mkr_lu=self.confl_mkr_lu_lose)

        self.assertEqual(self.img_lu, cs.get_or_upload_image_lu(
            self.tier, self.img_lu.name, self.vios_uuid, self.mock_stream_func,
            self.b_size))

        # I tried creating mine because his wasn't there at the start; and I
        # also created the image LU.
        self.assertEqual(2, self.mock_crt_lu.call_count)
        # Since I won, I did the upload
        self.assertEqual(1, self.mock_upload_lu.call_count)
        # I did remove my marker from the SSP
        self.mkr_lu.delete.assert_called_once_with()
        # I didn't remove the image LU (because I won)
        self.img_lu.delete.assert_not_called()
        # I never slept
        self.mock_sleep.assert_not_called()
        # I searched the first time through, and in _upload_conflict
        self.assertEqual(2, self.mock_luent_srch.call_count)
        # IRL, the other guy will have removed his marker LU at some point.
        # Here, we can expect it to remain, so there's one "extra".
        self.assertEqual(self.exp_num_lus + 1, len(self.entries))

    def test_crt_img_lu_raises(self):
        """Exception during crt_lu of the image LU."""
        self.setup_crt_lu_mock(IOError('crt_lu raises on the image LU'),
                               conflicting_mkr_lu=self.confl_mkr_lu_lose)

        self.assertRaises(IOError, cs.get_or_upload_image_lu, self.tier,
                          self.img_lu.name, self.vios_uuid,
                          self.mock_stream_func, self.b_size)

        # I didn't get to the upload
        self.mock_upload_lu.assert_not_called()
        # I never slept
        self.mock_sleep.assert_not_called()
        # I removed my marker
        self.mkr_lu.delete.assert_called_once_with()
        # I didn't remove the image LU (because I failed to create it)
        self.img_lu.delete.assert_not_called()
        # I searched the first time through, and in _upload_conflict
        self.assertEqual(2, self.mock_luent_srch.call_count)
        # We left the SSP as it was (plus the other guy's extra, which would
        # actually be removed normally).
        self.assertEqual(self.exp_num_lus, len(self.entries))

    def test_upload_raises(self):
        """I win; upload_lu raises after crt_lu of the image LU."""
        self.setup_crt_lu_mock(self.crt_img_lu,
                               conflicting_mkr_lu=self.confl_mkr_lu_lose)
        self.mock_upload_lu.side_effect = IOError('upload_lu raises.')

        self.assertRaises(IOError, cs.get_or_upload_image_lu, self.tier,
                          self.img_lu.name, self.vios_uuid,
                          self.mock_stream_func, self.b_size)

        # I created my marker and the image LU
        self.assertEqual(2, self.mock_crt_lu.call_count)
        # Since I won, I tried the upload
        self.assertEqual(1, self.mock_upload_lu.call_count)
        # I never slept
        self.mock_sleep.assert_not_called()
        # I removed both the real LU and my marker
        self.mkr_lu.delete.assert_called_once_with()
        self.img_lu.delete.assert_called_once_with()
        # I searched the first time through, and in _upload_conflict
        self.assertEqual(2, self.mock_luent_srch.call_count)
        # We left the SSP as it was (plus the other guy's extra, which would
        # actually be removed normally).
        self.assertEqual(self.exp_num_lus, len(self.entries))



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from pypowervm import const
from pypowervm.i18n import _
from pypowervm.wrappers import entry_wrapper
from pypowervm.wrappers import mtms as mtms_wrapper

LOG = logging.getLogger(__name__)


_POWER_ENTERPRISE_POOL = 'PowerEnterprisePool'
_POWER_ENTERPRISE_POOL_MEMBER = 'PowerEnterprisePoolMember'

_P_POOL_ID = 'PoolID'
_P_POOL_NAME = 'PoolName'

_P_COMPLIANCE_STATE = 'ComplianceState'
_P_COMPLIANCE_REMAINING_HOURS = 'ComplianceRemainingHours'
_PM_PROC_COMPLIANCE_HOURS_LEFT = 'ProcComplianceRemainingHours'
_PM_MEM_COMPLIANCE_HOURS_LEFT = 'MemComplianceRemainingHours'

_P_TOTAL_MOBILE_PROCS = 'TotalMobileCoDProcUnits'
_P_AVAIL_MOBILE_PROCS = 'AvailableMobileCoDProcUnits'
_UNRET_MOBILE_PROCS = 'UnreturnedMobileCoDProcUnits'
_P_TOTAL_MOBILE_MEM = 'TotalMobileCoDMemory'
_P_AVAIL_MOBILE_MEM = 'AvailableMobileCoDMemory'
_UNRET_MOBILE_MEM = 'UnreturnedMobileCoDMemory'

_PM_MOBILE_PROCS = 'MobileCoDProcUnits'
_PM_MOBILE_MEM = 'MobileCoDMemory'
_PM_INACTIVE_PROCS = 'InactiveProcUnits'
_PM_INACTIVE_MEM = 'InactiveMemory'

_PM_SYS_NAME = 'ManagedSystemName'
_PM_SYS_INSTALLED_PROCS = 'ManagedSystemInstalledProcUnits'
_PM_SYS_INSTALLED_MEM = 'ManagedSystemInstalledMemory'
_PM_SYS_MTMS = 'ManagedSystemMachineTypeModelSerialNumber'
_PM_SYS_STATE = 'ManagedSystemState'

_MGMT_CONSOLES = 'PowerEnterprisePoolManagementConsoles'
_MGMT_CONSOLE = 'PowerEnterprisePoolManagementConsole'
_MGMT_CONSOLE_NAME = 'ManagementConsoleName'
_MGMT_CONSOLE_MTMS = 'ManagementConsoleMachineTypeModelSerialNumber'
_MGMT_CONSOLE_IS_MASTER_CONSOLE = 'IsMasterConsole'
_MGMT_CONSOLE_IP_ADDR = 'ManagementConsoleIPAddress'


class ComplianceState(object):
    IN_COMPLIANCE = 'InCompliance'
    APPROACHING_OUT_OF_COMPLIANCE_SERVER = 'ApproachingOutOfComplianceServer'
    APPROACHING_OUT_OF_COMPLIANCE_POOL = 'ApproachingOutOfCompliancePool'
    OUT_OF_COMPLIANCE = 'OutOfCompliance'
    UNAVAILABLE = 'Unavailable'


@entry_wrapper.EntryWrapper.pvm_type(_POWER_ENTERPRISE_POOL, has_metadata=True)
class Pool(entry_wrapper.EntryWrapper):
    """Wraps the Pool entries."""

    @property
    def id(self):
        """Integer enterprise pool ID."""
        return self._get_val_int(_P_POOL_ID)

    @property
    def name(self):
        """The name of the enterprise pool."""
        return self._get_val_str(_P_POOL_NAME)

    @property
    def compliance_state(self):
        """The compliance state of the enterprise pool."""
        return self._get_val_str(_P_COMPLIANCE_STATE)

    @entry_wrapper.Wrapper.xag_property(const.XAG.ADV)
    def compliance_hours_left(self):
        """Integer num of hours until the pool is considered out of compliance.

        Return default of 0 if it is not found.
        """
        return self._get_val_int(_P_COMPLIANCE_REMAINING_HOURS, default=0)

    @property
    def total_mobile_procs(self):
        """Integer num of the total mobile CoD proc units in the pool."""
        return self._get_val_int(_P_TOTAL_MOBILE_PROCS)

    @property
    def avail_mobile_procs(self):
        """Integer num of the available mobile CoD proc units in the pool."""
        return self._get_val_int(_P_AVAIL_MOBILE_PROCS, default=0)

    @property
    def unret_mobile_procs(self):
        """Integer num of the unreturned mobile CoD proc units in the pool."""
        return self._get_val_int(_UNRET_MOBILE_PROCS)

    @property
    def total_mobile_mem(self):
        """Integer num of the total mobile CoD memory (GB) in the pool."""
        return self._get_val_int(_P_TOTAL_MOBILE_MEM)

    @property
    def avail_mobile_mem(self):
        """Integer num of the available mobile CoD memory (GB) in the pool."""
        return self._get_val_int(_P_AVAIL_MOBILE_MEM, default=0)

    @property
    def unret_mobile_mem(self):
        """Integer num of the unreturned mobile CoD memory (GB) in the pool."""
        return self._get_val_int(_UNRET_MOBILE_MEM)

    @property
    def mgmt_consoles(self):
        """Returns a WrapperElemList of PoolMgmtConsole's."""
        elem = self._find_or_seed(_MGMT_CONSOLES)
        return entry_wrapper.WrapperElemList(
            elem, PoolMgmtConsole)

    @property
    def master_console_mtms(self):
        """The master console MTMS (machine type, model, serial number)."""
        for console in self.mgmt_consoles:
            if console.is_master_console:
                return console.mtms

        LOG.error(_('Unable to determine master management console MTMS '
                    '(machine type, model, serial number) from '
                    '%(identifier)s because no %(param)s was marked as the '
                    'master console for the pool.') %
                  {'identifier': self._type_and_uuid,
                   'param': _MGMT_CONSOLE})

        return None


@entry_wrapper.EntryWrapper.pvm_type(_POWER_ENTERPRISE_POOL_MEMBER,
                                     has_metadata=True)
class PoolMember(entry_wrapper.EntryWrapper):
    """Wraps the PoolMember entries."""

    @property
    def mobile_procs(self):
        """Integer num of the mobile CoD proc units on the system."""
        return self._get_val_int(_PM_MOBILE_PROCS)

    @mobile_procs.setter
    def mobile_procs(self, value):
        self.set_parm_value(_PM_MOBILE_PROCS, value)

    @property
    def mobile_mem(self):
        """Integer amount of mobile CoD memory (GB) on the system."""
        return self._get_val_int(_PM_MOBILE_MEM)

    @mobile_mem.setter
    def mobile_mem(self, value):
        self.set_parm_value(_PM_MOBILE_MEM, value)

    @property
    def inactive_procs(self):
        """Integer num of the inactive (dark) proc units on the system."""
        return self._get_val_int(_PM_INACTIVE_PROCS)

    @property
    def inactive_mem(self):
        """Integer amount of inactive (dark) memory (GB) on the system."""
        return self._get_val_int(_PM_INACTIVE_MEM)

    @property
    def unret_mobile_procs(self):
        """Integer num of the unreturned mobile CoD proc units on the sys."""
        return self._get_val_int(_UNRET_MOBILE_PROCS)

    @property
    def unret_mobile_mem(self):
        """Integer amount of unreturned mobile CoD memory (GB) on the sys."""
        return self._get_val_int(_UNRET_MOBILE_MEM)

    @entry_wrapper.Wrapper.xag_property(const.XAG.ADV)
    def proc_compliance_hours_left(self):
        """Integer num of proc compliance hours remaining.

        Number of hours remaining until the system is considered out of
        compliance in terms of mobile procs.

        Return default of 0 if it is not found.
        """
        return self._get_val_int(_PM_PROC_COMPLIANCE_HOURS_LEFT, default=0)

    @entry_wrapper.Wrapper.xag_property(const.XAG.ADV)
    def mem_compliance_hours_left(self):
        """Integer num of memory compliance hours remaining.

        Number of hours remaining until the system is considered out of
        compliance in terms of mobile memory.

        Return default of 0 if it is not found.
        """
        return self._get_val_int(_PM_MEM_COMPLIANCE_HOURS_LEFT, default=0)

    @property
    def sys_name(self):
        """The name of the system that corresponds to this pool member."""
        return self._get_val_str(_PM_SYS_NAME)

    @property
    def sys_installed_procs(self):
        """Integer num of the installed proc units on the system."""
        return self._get_val_int(_PM_SYS_INSTALLED_PROCS)

    @property
    def sys_installed_mem(self):
        """Integer amount of installed memory (MB) on the system."""
        return self._get_val_int(_PM_SYS_INSTALLED_MEM)

    @property
    def sys_mtms(self):
        """The MTMS (machine type, model, serial number) of the system."""
        sys_mtms_element = self._find(_PM_SYS_MTMS)
        return mtms_wrapper.MTMS.wrap(sys_mtms_element)

    @property
    def sys_state(self):
        """The state of the system."""
        return self._get_val_str(_PM_SYS_STATE)

    @property
    def mgmt_consoles(self):
        """Returns a WrapperElemList of PoolMgmtConsole's."""
        elem = self._find_or_seed(_MGMT_CONSOLES)
        return entry_wrapper.WrapperElemList(
            elem, PoolMgmtConsole)


@entry_wrapper.ElementWrapper.pvm_type(_MGMT_CONSOLE,
                                       has_metadata=True)
class PoolMgmtConsole(entry_wrapper.ElementWrapper):
    """Wraps the PoolMgmtConsole elements."""

    @property
    def name(self):
        """String value for the name of the management console."""
        return self._get_val_str(_MGMT_CONSOLE_NAME)

    @property
    def mtms(self):
        """The MTMS (machine type, model, serial number) of the console."""
        return mtms_wrapper.MTMS.wrap(self.element.find(_MGMT_CONSOLE_MTMS))

    @property
    def is_master_console(self):
        """Boolean for whether or not this console is master for the pool."""
        return self._get_val_bool(_MGMT_CONSOLE_IS_MASTER_CONSOLE)

    @property
    def ip_addr(self):
        """String value for the IP address of the console."""
        return self._get_val_str(_MGMT_CONSOLE_IP_ADDR)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Wrappers and related artifacts around /rest/api/uom/Event.

Events are produced semi-synchronously by the REST server.  Event.get() will
return right away if there are events waiting to be retrieved.  Otherwise, it
will block for up to ten seconds.  If no events have been produced in that
time, the request will return an empty feed.  Using this mechanism, it is
practical to poll Event.get in a hard loop with no sleeping on the client side.

An Event has a type (Event.etype), a data field (Event.data), and a detail
field (Event.detail).  For *_URI types, the data field contains the REST URI of
the object triggering the event, and the detail field may provide more granular
information about the event.

Events arrive in the order they are produced.  Special event types
CACHE_CLEARED and MISSING_EVENTS indicate that the client should refetch any
objects of interest before processing subsequent events.

Different clients may access the same Event feed at the same time.  The REST
server keeps track of which client has seen which events based on an
application ID (the `appid` argument to the get method).  Requests using a
given application ID for repeated requests will receive a single stream of
events (no duplicates).  Two clients using different application IDs will each
receive the same stream of events.

It is possible to push a custom event to the server.  This event will appear to
all active listeners as a CUSTOM_CLIENT_EVENT.  To use this mechanism,
construct the event with Event.bld(), supplying any values desired in data and
detail, and invoke .create() on the resulting Event wrapper.
"""

import pypowervm.wrappers.entry_wrapper as ewrap

# Constants for Event
_E_SCHEMA_TYPE = 'Event'
_E_TYPE = 'EventType'
_E_ID = 'EventID'
_E_DATA = 'EventData'
_E_DETAIL = 'EventDetail'

_E_EL_ORDER = (_E_TYPE, _E_ID, _E_DATA, _E_DETAIL)


class EventType(object):
    """Enumeration of event types (from EventType.Enum)."""
    INVALID_URI = 'INVALID_URI'
    CACHE_CLEARED = 'CACHE_CLEARED'
    MISSING_EVENTS = 'MISSING_EVENTS'
    ADD_URI = 'ADD_URI'
    MODIFY_URI = 'MODIFY_URI'
    DELETE_URI = 'DELETE_URI'
    NEW_CLIENT = 'NEW_CLIENT'
    HIDDEN_URI = 'HIDDEN_URI'
    VISIBLE_URI = 'VISIBLE_URI'
    CUSTOM_CLIENT_EVENT = 'CUSTOM_CLIENT_EVENT'


@ewrap.EntryWrapper.pvm_type(_E_SCHEMA_TYPE, child_order=_E_EL_ORDER)
class Event(ewrap.EntryWrapper):

    @classmethod
    def get(cls, adapter, appid):
        """Retrieve the latest Event feed for a given application ID.

        Note: This request may block for a finite amount of time (on the order
        of 10s) while the server is waiting for new events to occur.

        :param adapter: pypowervm.adapter.Adapter for REST API communication.
        :param appid: A hex string identifying the unique consumer.  Consumers
                      pulling Event feeds will see the same events duplicated
                      in each request that uses a different appid.  To see a
                      single stream of unique events, a consumer should make
                      repeated requests with the same appid.
        :return: Feed of Event EntryWrapper objects (may be empty).
        """
        return super(Event, cls).get(adapter, xag=[], add_qp=[
            ('QUEUE_CLIENTKEY_METHOD', 'USE_APPLICATIONID'),
            ('QUEUE_APPLICATIONID', appid)])

    @classmethod
    def bld(cls, adapter, data, detail):
        """Construct a custom Event.

        Invoke .create() on the resulting Event to broadcast it to active
        listeners.

        :param adapter: pypowervm.adapter.Adapter for REST API communication.
        :param data: Any desired string to be included in the 'data' field of
                     the Event.  May be None.
        :param detail: Any desired string to be included in the 'detail' field
                       of the Event.  May be None.
        :return: An Event wrapper suitable for sending to the REST server via
                 the .create() method.
        """
        event = cls._bld(adapter)
        if data is not None:
            event.set_parm_value(_E_DATA, data)
        if detail is not None:
            event.set_parm_value(_E_DETAIL, detail)
        return event

    @property
    def etype(self):
        """The Event type, one of the EventType enum values."""
        return self._get_val_str(_E_TYPE)

    @property
    def eid(self):
        """Unique sequence identifier of this Event."""
        return self._get_val_str(_E_ID)

    @property
    def data(self):
        """Event data; for *_URI EventType, the URI of the affected object."""
        return self._get_val_str(_E_DATA)

    @property
    def detail(self):
        """Custom Event detail; semantics dependent on type & data."""
        return self._get_val_str(_E_DETAIL)

    def __str__(self):
        return "Event(id=%s, type=%s, data=%s, detail=%s)" % (
            self.eid, self.etype, self.data, self.detail)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

COMMENT = "#"
SECTION = "SECTION:"
END_OF_SECTION = "END OF SECTION"


def load_xml_sections(file_name):
    """Loads a file that contains xml sections

    This method takes a file that contains xml sections and returns
    a dict of them.  It's useful for testing the generation of the sections.

    See ../data/lpar_sections.txt for an example of the file contents.

    :param file_name: The name of the file to load.
    """
    def _read_section(nm):
        buf = ""
        while True:
            ln = sect_file.readline()
            if ln is None or len(ln) == 0:
                raise Exception("Could not find end of section %s of file %s" %
                                (nm, file_name))

            if ln.startswith(END_OF_SECTION):
                return buf

            buf += ln.strip('\n')

        return buf

    sections = {}

    # First try to load the name as passed in.
    dirname = os.path.dirname(file_name)
    if dirname is None or dirname == '':
        dirname = os.path.dirname(os.path.dirname(__file__))
        file_name = os.path.join(dirname, "data", file_name)

    sect_file = open(file_name, "r")

    if sect_file is None:
        raise Exception("Could not load %s" % file_name)

    while True:
        line = sect_file.readline()
        if line is None or len(line) == 0:
            break

        if len(line.strip()) == 0:
            continue

        if line.startswith(COMMENT):
            continue

        if line.startswith(SECTION):
            # Get the name of the section
            name = line[len(SECTION):].strip()
            # Get the data
            data = _read_section(name)
            sections[name] = data
        else:
            sect_file.close()
            raise Exception("Unknown line in file %s: %s" %
                            (file_name, line))

    sect_file.close()

    return sections



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import mock
import six
import testtools

from pypowervm.tests import test_fixtures as fx
from pypowervm.tests.test_utils import xml_sections
from pypowervm.utils import lpar_builder as lpar_bldr
import pypowervm.utils.uuid as pvm_uuid
from pypowervm.wrappers import base_partition as bp
from pypowervm.wrappers import logical_partition as lpar

LPAR_BLDR_DATA = 'lpar_builder.txt'


class TestLPARBuilder(testtools.TestCase):
    """Unit tests for the lpar builder."""

    def setUp(self):
        super(TestLPARBuilder, self).setUp()
        self.sections = xml_sections.load_xml_sections(LPAR_BLDR_DATA)
        self.adpt = self.useFixture(fx.AdapterFx()).adpt

        def _bld_mgd_sys(proc_units, mem_reg, srr, pcm, ame):
            # Build a fake managed system wrapper
            mngd_sys = mock.Mock()
            type(mngd_sys).proc_units_avail = (
                mock.PropertyMock(return_value=proc_units))
            type(mngd_sys).memory_region_size = (
                mock.PropertyMock(return_value=mem_reg))
            capabilities = {
                'simplified_remote_restart_capable': srr,
                'ibmi_restrictedio_capable': True,
                'active_memory_expansion_capable': ame
            }
            mngd_sys.get_capabilities.return_value = capabilities
            type(mngd_sys).proc_compat_modes = (
                mock.PropertyMock(return_value=pcm))
            return mngd_sys

        self.mngd_sys = _bld_mgd_sys(20.0, 128, True,
                                     bp.LPARCompat.ALL_VALUES, False)
        self.mngd_sys_no_srr = _bld_mgd_sys(20.0, 128, False, ['POWER6'],
                                            False)
        self.mngd_sys_ame = _bld_mgd_sys(20.0, 128, True,
                                         bp.LPARCompat.ALL_VALUES, True)
        self.stdz_sys1 = lpar_bldr.DefaultStandardize(self.mngd_sys)
        self.stdz_sys2 = lpar_bldr.DefaultStandardize(self.mngd_sys_no_srr)
        self.stdz_sys3 = lpar_bldr.DefaultStandardize(self.mngd_sys_ame)

    def assert_xml(self, entry, string):
        self.assertEqual(entry.element.toxmlstring(),
                         six.b(string.rstrip('\n')))

    def test_proc_modes(self):
        # Base minimum attrs
        attr = dict(name='TheName', memory=1024, vcpu=1)
        # No proc keys specified
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertFalse(bldr._shared_proc_keys_specified())
        self.assertFalse(bldr._dedicated_proc_keys_specified())
        # Default is shared
        self.assertTrue(bldr._shared_procs_specified())
        # Shared proc keys specified
        for key in ('proc_units', 'max_proc_units', 'min_proc_units',
                    'uncapped_weight'):
            bldr = lpar_bldr.LPARBuilder(self.adpt, dict(attr, **{key: 1}),
                                         self.stdz_sys1)
            self.assertTrue(bldr._shared_proc_keys_specified())
            self.assertFalse(bldr._dedicated_proc_keys_specified())
            self.assertTrue(bldr._shared_procs_specified())
        # Shared modes specified
        for mode in ('capped', 'uncapped'):
            bldr = lpar_bldr.LPARBuilder(
                self.adpt, dict(attr, sharing_mode=mode), self.stdz_sys1)
            self.assertTrue(bldr._shared_proc_keys_specified())
            self.assertFalse(bldr._dedicated_proc_keys_specified())
            self.assertTrue(bldr._shared_procs_specified())
        # Dedicated modes specified
        for mode in ('sre idle proces', 'sre idle procs active',
                     'sre idle procs always', 'keep idle procs'):
            bldr = lpar_bldr.LPARBuilder(
                self.adpt, dict(attr, sharing_mode=mode), self.stdz_sys1)
            self.assertFalse(bldr._shared_proc_keys_specified())
            self.assertTrue(bldr._dedicated_proc_keys_specified())
            self.assertFalse(bldr._shared_procs_specified())
        # Dedicated proc explicitly true
        bldr = lpar_bldr.LPARBuilder(
            self.adpt, dict(attr, dedicated_proc='TRUE'), self.stdz_sys1)
        self.assertFalse(bldr._shared_proc_keys_specified())
        self.assertFalse(bldr._dedicated_proc_keys_specified())
        self.assertFalse(bldr._shared_procs_specified())
        # Dedicated proc explicitly false
        bldr = lpar_bldr.LPARBuilder(
            self.adpt, dict(attr, dedicated_proc='NO'), self.stdz_sys1)
        self.assertFalse(bldr._shared_proc_keys_specified())
        self.assertFalse(bldr._dedicated_proc_keys_specified())
        self.assertTrue(bldr._shared_procs_specified())

    def test_builder(self):
        # Build the minimum attributes, Shared Procs
        attr = dict(name='TheName', env=bp.LPARType.AIXLINUX, memory=1024,
                    vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertIsNotNone(bldr)

        new_lpar = bldr.build()
        self.assertIsNotNone(new_lpar)
        self.assert_xml(new_lpar, self.sections['shared_lpar'])
        self.assertEqual('TheName', new_lpar.name)

        # Rebuild the same lpar with a different name
        attr['name'] = 'NewName'
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        rbld_lpar = bldr.rebuild(new_lpar)
        self.assertEqual('NewName', rbld_lpar.name)

        # Build the minimum attributes, Dedicated Procs
        attr = dict(name='TheName', env=bp.LPARType.AIXLINUX, memory=1024,
                    vcpu=1, dedicated_proc=True)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertIsNotNone(bldr)

        new_lpar = bldr.build()
        self.assertIsNotNone(new_lpar)
        self.assert_xml(new_lpar.entry, self.sections['dedicated_lpar'])

        # Build the minimum attributes, Dedicated Procs = 'true'
        attr = dict(name='TheName', env=bp.LPARType.AIXLINUX, memory=1024,
                    vcpu=1, dedicated_proc='true')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assert_xml(new_lpar.entry, self.sections['dedicated_lpar'])

        # Leave out memory
        attr = dict(name=lpar, env=bp.LPARType.AIXLINUX, vcpu=1)
        self.assertRaises(
            lpar_bldr.LPARBuilderException, lpar_bldr.LPARBuilder, self.adpt,
            attr, self.stdz_sys1)

        # Bad memory lmb multiple
        attr = dict(name='lpar', memory=3333, env=bp.LPARType.AIXLINUX, vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Check the validation of the LPAR type when not specified
        attr = dict(name='TheName', memory=1024, vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assert_xml(new_lpar, self.sections['shared_lpar'])

        # LPAR name too long
        attr = dict(name='lparlparlparlparlparlparlparlparlparlparlparlpar'
                    'lparlparlparlparlparlparlparlparlparlparlparlparlparlpar',
                    memory=1024, env=bp.LPARType.AIXLINUX, vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(lpar_bldr.LPARBuilderException, bldr.build)

        # Test setting uuid
        uuid1 = pvm_uuid.convert_uuid_to_pvm(str(uuid.uuid4()))
        attr = dict(name='lpar', memory=1024, uuid=uuid1, vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        lpar_w = bldr.build()
        self.assertEqual(uuid1.upper(), lpar_w.uuid)

        # Test setting id
        id1 = 1234
        attr = dict(name='lpar', memory=1024, uuid=uuid1, vcpu=1, id=id1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        lpar_w = bldr.build()
        self.assertEqual(id1, lpar_w.id)

        # Bad LPAR type
        attr = dict(name='lpar', memory=1024, env='BADLPARType', vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Bad IO Slots
        attr = dict(name='lpar', memory=1024, max_io_slots=0,
                    env=bp.LPARType.AIXLINUX, vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        attr = dict(name='lpar', memory=1024, max_io_slots=(65534+1),
                    env=bp.LPARType.AIXLINUX, vcpu=1)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Good non-defaulted IO Slots and SRR
        attr = dict(name='TheName', memory=1024, max_io_slots=64,
                    env=bp.LPARType.AIXLINUX, vcpu=1, srr_capability=False)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assert_xml(new_lpar, self.sections['shared_lpar'])

        # Bad SRR value.
        attr = dict(name='lpar', memory=1024, max_io_slots=64,
                    env=bp.LPARType.AIXLINUX, vcpu=1, srr_capability='Frog')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Uncapped / capped shared procs
        attr = dict(name='TheName', env=bp.LPARType.AIXLINUX, memory=1024,
                    vcpu=1, sharing_mode=bp.SharingMode.CAPPED,
                    srr_capability='true')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assert_xml(new_lpar, self.sections['capped_lpar'])

        # Uncapped and no SRR capability
        attr = dict(name='TheName', env=bp.LPARType.AIXLINUX, memory=1024,
                    vcpu=1, sharing_mode=bp.SharingMode.UNCAPPED,
                    uncapped_weight=100, processor_compatibility='POWER6')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys2)
        new_lpar = bldr.build()
        self.assert_xml(new_lpar, self.sections['uncapped_lpar'])

        # Build dedicated but only via dedicated attributes
        m = bp.DedicatedSharingMode.SHARE_IDLE_PROCS_ALWAYS
        attr = dict(name='TheName', env=bp.LPARType.AIXLINUX, memory=1024,
                    vcpu=1, sharing_mode=m, processor_compatibility='PoWeR7')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assert_xml(new_lpar.entry,
                        self.sections['ded_lpar_sre_idle_procs_always'])

        # Desired mem outside min
        attr = dict(name='lpar', memory=1024, env=bp.LPARType.AIXLINUX, vcpu=1,
                    min_mem=2048)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Desired mem outside max
        attr = dict(name='lpar', memory=5000, env=bp.LPARType.AIXLINUX, vcpu=1,
                    max_mem=2048)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # AME not supported on host
        attr = dict(name='lpar', memory=1024, env=bp.LPARType.AIXLINUX, vcpu=1,
                    ame_factor='1.5')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # AME outside valid range
        attr = dict(name='lpar', memory=1024, env=bp.LPARType.AIXLINUX, vcpu=1,
                    ame_factor='0.5')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys3)
        self.assertRaises(ValueError, bldr.build)

        # Desired vcpu outside min
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=1,
                    min_vcpu=2)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Desired vcpu outside max
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    max_vcpu=2)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Ensure the calculated procs are not below the min
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    min_proc_units=3)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        procs = new_lpar.proc_config.shared_proc_cfg
        self.assertEqual(3.0, procs.min_units)

        # Ensure the calculated procs are all 0.5
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=1,
                    proc_units=0.5)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        procs = new_lpar.proc_config.shared_proc_cfg
        self.assertEqual(0.5, procs.min_units)
        self.assertEqual(0.5, procs.max_units)
        self.assertEqual(0.5, procs.desired_units)

        # Create a temp standardizer with a smaller proc units factor
        stdz = lpar_bldr.DefaultStandardize(self.mngd_sys,
                                            proc_units_factor=0.1)
        # Ensure the min, max, and desired proc units works as VCPU is scaled.
        for x in [1, 5, 10, 17, 20]:
            attr = dict(name='lpar', memory=2048, vcpu=x)
            bldr = lpar_bldr.LPARBuilder(self.adpt, attr, stdz)
            new_lpar = bldr.build()
            procs = new_lpar.proc_config.shared_proc_cfg
            self.assertEqual(round(0.1 * x, 2), procs.min_units)
            self.assertEqual(round(0.1 * x, 2), procs.max_units)
            self.assertEqual(round(0.1 * x, 2), procs.desired_units)

        # Ensure the calculated procs are below the max
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    max_proc_units=2.1)
        stdz = lpar_bldr.DefaultStandardize(
            self.mngd_sys, proc_units_factor=0.9)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, stdz)
        new_lpar = bldr.build()
        procs = new_lpar.proc_config.shared_proc_cfg
        self.assertEqual(2.1, procs.max_units)

        # Ensure proc units factor is between 0.1 and 1.0
        self.assertRaises(
            lpar_bldr.LPARBuilderException,
            lpar_bldr.DefaultStandardize,
            self.mngd_sys, proc_units_factor=1.01)
        self.assertRaises(
            lpar_bldr.LPARBuilderException,
            lpar_bldr.DefaultStandardize,
            self.mngd_sys, proc_units_factor=0.01)

        # Avail priority outside max
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    avail_priority=332)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Avail priority bad parm
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    avail_priority='BADVALUE')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertRaises(ValueError, bldr.build)

        # Avail priority at min value
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    avail_priority=0)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assertEqual(new_lpar.avail_priority, 0)

        # Avail priority at max value
        attr = dict(name='lpar', memory=2048, env=bp.LPARType.AIXLINUX, vcpu=3,
                    avail_priority=255)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        new_lpar = bldr.build()
        self.assertEqual(new_lpar.avail_priority, 255)

        # Proc compat
        for pc in bp.LPARCompat.ALL_VALUES:
            attr = dict(name='name', memory=1024, vcpu=1,
                        processor_compatibility=pc)
            bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
            new_lpar = bldr.build()
            self.assertEqual(new_lpar.pending_proc_compat_mode, pc)

        attr = dict(name='name', memory=1024, vcpu=1,
                    processor_compatibility='POWER6')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys2)
        new_lpar = bldr.build()
        self.assertEqual(new_lpar.pending_proc_compat_mode, 'POWER6')

        # Ensure failure occurs on validation after the host supported
        # proc modes are loaded and not on convert_value which converts
        # across all acceptable proc modes.
        # This works because 'POWER8' is in LPARCompat.ALL_VALUES
        attr = dict(name='name', memory=1024, vcpu=1,
                    processor_compatibility='POWER8')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys2)
        exp_msg = ("Value 'POWER8' is not valid for field 'Processor "
                   "Compatability Mode' with acceptable choices: ['POWER6']")
        try:
            bldr.build()
        except Exception as e:
            self.assertEqual(six.text_type(e), exp_msg)

        # Build a VIOS
        attr = dict(name='TheName', env=bp.LPARType.VIOS, memory=1024,
                    vcpu=1, dedicated_proc=True)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertIsNotNone(bldr)

        new_lpar = bldr.build()
        self.assertIsNotNone(new_lpar)
        self.assert_xml(new_lpar.entry, self.sections['vios'])

    def test_IBMi(self):
        attr = dict(name='TheName', env=bp.LPARType.OS400, memory=1024,
                    vcpu=1, ame_factor=False)
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertIsNotNone(bldr)

        new_lpar = bldr.build()
        self.assertIsNotNone(new_lpar)
        self.assertTrue(new_lpar.restrictedio)
        tag_io = new_lpar.io_config.tagged_io
        self.assertEqual('HMC', tag_io.console)
        self.assertEqual('0', tag_io.load_src)
        self.assertEqual('NONE', tag_io.alt_load_src)

        attr = dict(name='OS400LPAR', env=bp.LPARType.OS400, memory=1024,
                    vcpu=1, console='CONSOLE', load_src='9', alt_load_src='9')
        bldr = lpar_bldr.LPARBuilder(self.adpt, attr, self.stdz_sys1)
        self.assertIsNotNone(bldr)

        new_lpar = bldr.build()
        self.assertIsNotNone(new_lpar)
        self.assertTrue(new_lpar.restrictedio)
        tag_io = new_lpar.io_config.tagged_io
        self.assertEqual('CONSOLE', tag_io.console)
        self.assertEqual('9', tag_io.load_src)
        self.assertEqual('9', tag_io.alt_load_src)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import gc
import mock
import os
import requests.models as req_mod
import requests.structures as req_struct
import six
import subunit
import testtools

import pypowervm.adapter as adp
import pypowervm.exceptions as pvmex
import pypowervm.tests.lib as testlib
import pypowervm.tests.test_fixtures as fx

_logon_response_password = testlib.file2b("logon.xml")
_logon_response_file = testlib.file2b("logon_file.xml")


class TestSession(subunit.IsolatedTestCase, testtools.TestCase):
    """Test cases to test the Session classes and methods."""

    @mock.patch('time.sleep')
    @mock.patch('lxml.etree.fromstring', new=mock.Mock())
    @mock.patch('pypowervm.adapter.Session._get_auth_tok_from_file',
                new=mock.Mock())
    def test_Session(self, mock_sleep):
        """Ensure Session can be instantiated, and test logon retries."""
        # Passing in 0.0.0.0 will raise a ConnectionError or SSLError, but only
        # if it gets past all the __init__ setup since _logon is the last
        # statement.
        self.assertRaises((pvmex.ConnectionError, pvmex.SSLError), adp.Session,
                          '0.0.0.0', 'uid', 'pwd')
        mock_sleep.assert_not_called()
        # Now set up a retry
        self.assertRaises((pvmex.ConnectionError, pvmex.SSLError), adp.Session,
                          '0.0.0.0', 'uid', 'pwd', conn_tries=5)
        # 5 tries = 4 sleeps
        mock_sleep.assert_has_calls([mock.call(2)] * 4)
        # Ensure 404 on the logon URI also retries
        mock_sleep.reset_mock()
        with mock.patch('requests.Session.request') as mock_rq:
            mock_rq.side_effect = [mock.Mock(status_code=404),
                                   mock.Mock(status_code=204)]
            adp.Session(conn_tries=5)
            # Only retried once, after the 404
            mock_sleep.assert_called_once_with(2)

    @mock.patch('pypowervm.adapter.Session._logon', new=mock.Mock())
    @mock.patch('pypowervm.adapter._EventListener._get_events')
    def test_session_init(self, mock_get_evts):
        """Ensure proper parameter handling in the Session initializer."""
        mock_get_evts.return_value = {'general': 'init'}, [], []
        logfx = self.useFixture(fx.LoggingFx())
        # No params - local, file-based, http.
        sess = adp.Session()
        self.assertTrue(sess.use_file_auth)
        self.assertIsNone(sess.password)
        self.assertTrue(sess.username.startswith('pypowervm_'))
        self.assertEqual('localhost', sess.host)
        self.assertEqual('http', sess.protocol)
        self.assertEqual(12080, sess.port)
        self.assertEqual('http://localhost:12080', sess.dest)
        self.assertEqual(1200, sess.timeout)
        self.assertEqual('/etc/ssl/certs/', sess.certpath)
        self.assertEqual('.crt', sess.certext)
        # localhost + http is okay
        self.assertEqual(0, logfx.patchers['warning'].mock.call_count)

        # Verify unique session names
        sess2 = adp.Session()
        self.assertNotEqual(sess.username, sess2.username)

        # Ensure proper protocol, port, and certpath defaulting when remote
        sess = adp.Session(host='host', username='user', password='pass')
        self.assertFalse(sess.use_file_auth)
        self.assertIsNotNone(sess.password)
        self.assertEqual('user', sess.username)
        self.assertEqual('host', sess.host)
        self.assertEqual('https', sess.protocol)
        self.assertEqual(12443, sess.port)
        self.assertEqual('https://host:12443', sess.dest)
        self.assertEqual(1200, sess.timeout)
        self.assertEqual('/etc/ssl/certs/', sess.certpath)
        self.assertEqual('.crt', sess.certext)
        # non-localhost + (implied) https is okay
        self.assertEqual(0, logfx.patchers['warning'].mock.call_count)

    @mock.patch('pypowervm.adapter.Session._logon', new=mock.Mock())
    @mock.patch('pypowervm.adapter._EventListener._get_events')
    @mock.patch('imp.load_source')
    def test_session_ext_cfg(self, mock_load, mock_get_evts):
        """Test Session init with external config from env var."""
        mock_get_evts.return_value = {'general': 'init'}, [], []
        with mock.patch.dict(os.environ, {'PYPOWERVM_SESSION_CONFIG': 'path'}):
            sess = adp.Session()
        mock_load.assert_called_once_with('sesscfg', 'path')
        mock_load.return_value.session_config.assert_called_once_with(sess)

    @mock.patch('pypowervm.adapter.Session._logon')
    def test_session_init_remote_http(self, mock_logon):
        # Proper port defaulting and warning emitted when remote + http
        with self.assertLogs(adp.__name__, 'WARNING'):
            sess = adp.Session(host='host', protocol='http')
        self.assertEqual(12080, sess.port)

    @mock.patch.object(adp.Session, '_logon')
    @mock.patch.object(adp.Session, '_logoff')
    def test_session_clone(self, mock_logoff, mock_logon):

        sess = adp.Session()
        # Ensure the id that created the object is recorded.
        self.assertTrue(hasattr(sess, '_init_by'))

        # Create a shallow clone and ensure the _init_by does not match the id
        sess_clone = copy.copy(sess)
        self.assertTrue(hasattr(sess_clone, '_init_by'))
        self.assertNotEqual(sess._init_by, id(sess_clone))

        # Now test what happens when the clone is garbage collected.
        self.assertFalse(mock_logoff.called)
        sess_clone = None
        gc.collect()
        # The clone was not logged off
        self.assertFalse(mock_logoff.called)

        if six.PY2:
            # Ensure deep copies raise an exception.
            self.assertRaises(TypeError, copy.deepcopy, sess)
        else:
            # Or if works, it is not logged off
            sess_deepclone = copy.deepcopy(sess)
            # Make pep8 happy, use the clone
            self.assertIsNotNone(sess_deepclone)
            sess_deepclone = None
            gc.collect()
            # The clone was not logged off
            self.assertFalse(mock_logoff.called)

        sess = None
        gc.collect()
        # The original session was logged off
        self.assertTrue(mock_logoff.called)

    @mock.patch('pypowervm.util.validate_certificate')
    @mock.patch('requests.Session')
    def test_logon(self, mock_session, mock_validate_cert):
        """Ensure a Session can be created and log on to PowerVM."""

        # Init test data
        host = '0.0.0.0'
        user = 'user'
        pwd = 'pwd'
        auditmemento = 'audit'

        # Create a Response object, that will serve as a mock return value
        my_response = req_mod.Response()
        my_response.status_code = 200
        my_response.reason = 'OK'
        dict_headers = {'content-length': '576',
                        'x-powered-by': 'Servlet/3.0',
                        'set-cookie': 'JSESSIONID=0000a41BnJsGTNQvBGERA3wR1nj:'
                                      '759878cb-4f9a-4b05-a09a-3357abfea3b4; P'
                                      'ath=/; Secure; HttpOnly, CCFWSESSION=E4'
                                      'C0FFBE9130431DBF1864171ECC6A6E; Path=/;'
                                      ' Secure; HttpOnly',
                        'expires': 'Thu, 01 Dec 1994 16:00:00 GMT',
                        'x-transaction-id': 'XT10000073',
                        'cache-control': 'no-cache="set-cookie, set-cookie2"',
                        'date': 'Wed, 23 Jul 2014 21:51:10 GMT',
                        'content-type': 'application/vnd.ibm.powervm.web+xml; '
                                        'type=LogonResponse'}
        my_response.headers = req_struct.CaseInsensitiveDict(dict_headers)
        my_response._content = _logon_response_password

        # Mock out the method and class we are not currently testing
        session = mock_session.return_value
        session.request.return_value = my_response

        # Run the actual test
        result = adp.Session(host, user, pwd, auditmemento=auditmemento)

        # Verify the result
        self.assertTrue(result._logged_in)
        self.assertEqual('PUIoR6x0kP6fQqA7qZ8sLZQJ8MLx9JHfLCYzT4oGFSE2WaGIhaFX'
                         'IyQYvbqdKNS8QagjBpPi9NP7YR_h61SOJ3krS_RvKAp-oCf2p8x8'
                         'uvQrrDv-dUzc17IT5DkR7_jv2qc8iUD7DJ6Rw53a17rY0p63KqPg'
                         '9oUGd6Bn3fNDLiEwaBR4WICftVxUFj-tfWMOyZZY2hWEtN2K8ScX'
                         'vyFMe-w3SleyRbGnlR34jb0A99s=', result._sessToken)
        self.assertEqual(1, mock_validate_cert.call_count)
        # No X-MC-Type header => 'HMC' is assumed.
        self.assertEqual('HMC', result.mc_type)

        # Now test file-based authentication and X-MC-Type
        my_response._content = _logon_response_file

        # Local/HMC is bad
        self.assertRaises(pvmex.Error, adp.Session)

        my_response.headers['X-MC-Type'] = 'PVM'
        result = adp.Session()

        # Verify the result.
        self.assertTrue(result._logged_in)
        # Token read from token_file, as indicated by logon_file.xml response.
        self.assertEqual('file-based-auth-token', result._sessToken)
        # validate_certificate should not have been called again
        self.assertEqual(1, mock_validate_cert.call_count)
        self.assertEqual('PVM', result.mc_type)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock
import testtools

from pypowervm import exceptions as pvm_exc
from pypowervm.tasks import ibmi
import pypowervm.tests.tasks.util as tju
import pypowervm.tests.test_fixtures as pvm_fx
from pypowervm.wrappers import virtual_io_server as pvm_vios

VIOS_FEED = 'fake_vios_feed.txt'


class TestIBMi(testtools.TestCase):
    """Unit Tests for IBMi changes."""

    def setUp(self, traits_type):
        super(TestIBMi, self).setUp()
        self.traits = traits_type
        self.apt = self.useFixture(pvm_fx.AdapterFx(
            traits=self.traits)).adpt
        self.vio_feed = pvm_vios.VIOS.wrap(
            tju.load_file(VIOS_FEED, self.apt))
        self.vioslist = [self.vio_feed[0], self.vio_feed[1]]

    @staticmethod
    def _validate_settings(self, boot_type, traits_type, entry):
        self.assertEqual('b', entry.desig_ipl_src)
        self.assertEqual('normal', entry.keylock_pos)

        if traits_type == pvm_fx.LocalPVMTraits:
            self.assertEqual('HMC', entry.io_config.tagged_io.console)
        else:
            self.assertEqual('HMC', entry.io_config.tagged_io.console)
        if boot_type == 'npiv':
            self.assertEqual('3', entry.io_config.tagged_io.load_src)
            self.assertEqual('4', entry.io_config.tagged_io.alt_load_src)
        else:
            self.assertEqual('2', entry.io_config.tagged_io.load_src)
            self.assertEqual('2', entry.io_config.tagged_io.alt_load_src)

    def _validate_ibmi_settings(self, mock_viosw):
        mock_viosw.return_value = self.vioslist
        mock_lparw = mock.MagicMock()
        mock_lparw.id = 22

        # Test update load source with npiv boot
        boot_type = 'npiv'
        entry = ibmi.update_ibmi_settings(self.apt, mock_lparw, boot_type)
        self._validate_settings(self, boot_type, self.traits, entry)

        # Test update load source with vscsi boot
        boot_type = 'vscsi'
        entry = ibmi.update_ibmi_settings(self.apt, mock_lparw, boot_type)
        self._validate_settings(self, boot_type, self.traits, entry)

        # Test bad path if load source is not found
        mock_lparw.reset_mock()
        mock_lparw.id = 220
        boot_type = 'vscsi'
        self.assertRaises(pvm_exc.IBMiLoadSourceNotFound,
                          ibmi.update_ibmi_settings, self.apt, mock_lparw,
                          boot_type)


class TestIBMiWithHMC(TestIBMi):
    """Unit Tests for IBMi changes for HMC."""

    def setUp(self):
        super(TestIBMiWithHMC, self).setUp(pvm_fx.RemoteHMCTraits)

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.wrap')
    def test_update_ibmi_settings(self, mock_viosw):
        self._validate_ibmi_settings(mock_viosw)


class TestIBMiWithPVM(TestIBMi):
    """Unit Tests for IBMi changes for PVM."""

    def setUp(self):
        super(TestIBMiWithPVM, self).setUp(pvm_fx.LocalPVMTraits)

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.wrap')
    def test_update_ibmi_settings(self, mock_viosw):
        self._validate_ibmi_settings(mock_viosw)

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.wrap')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VStorageMapping.'
                'client_adapter', new_callable=mock.PropertyMock,
                return_value=None)
    def test_update_ibmi_settings_w_stale_adapters(self, mock_c_adap,
                                                   mock_viosw):
        mock_lparw = mock.MagicMock()
        mock_lparw.id = 22
        self.assertRaises(pvm_exc.IBMiLoadSourceNotFound,
                          ibmi.update_ibmi_settings, self.apt,
                          mock_lparw, 'vscsi')



# Copyright 2015, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Identify behavioral traits specific to a PowerVM server type/version."""

import weakref


class APITraits(object):
    """Represents capabilities inherent to the backing API server.

    For instance, depending on a schema version, or the backing REST server,
    there may be different code paths that the user needs to go through.

    A key example of this would be the VirtualNetworks.  The HMC requires
    that Virtual Networks be the driver of code paths for various code paths.
    However, in other API servers, the Virtual Networks are simply realized
    based off the VLANs/virtual switches that adapters are currently tied
    to.

    This class encapsulates the various traits so that tasks and users do not
    have to inspect the header data directly to determine how the API should
    behave.
    """

    def __init__(self, session):
        # Avoid circular references to the session by using a weak reference.
        # Circular references prevent garabage collection.
        self._sess_ref = weakref.ref(session)
        self._is_hmc = session.mc_type == 'HMC'

    @property
    def session(self):
        # Get the session through the weak reference
        return self._sess_ref()

    @property
    def vnet_aware(self):
        """Indicates whether Virtual Networks are pre-reqs to Network changes.

        Some APIs (such as modifying the SEA or creating a Client Network
        Adapter) require that the VirtualNetwork (or VNet wrapper) be
        pre-created for the operation.  This is typically done when working
        against an HMC.

        This trait will return True if the Virtual Networks need to be
        passed in on NetworkBridge or Client Network Adapter creation, or
        False if the API should directly work with VLANs and Virtual Switches.
        """
        return self._is_hmc

    @property
    def has_lpar_profiles(self):
        """Indicates whether the platform manager supports LPAR profiles.

        This trait will return True if LPAR profiles are supported.
        """
        return self._is_hmc

    @property
    def local_api(self):
        """Indicates whether or not the PowerVM API Server is running locally.

        The PowerVM API server in some deployments may be running co-located
        with the pypowervm API.  In those cases, certain optimizations may be
        available (like uploading from a file instead of a pipe).

        This trait is a coarse check to determine, for certain, if the API
        is co-located on the same server.
        """
        # If the file auth is set to true, we must be colocated.  All other
        # routes could be error prone and lead to significant branches of
        # complexity.
        return self.session.use_file_auth

    @property
    def dynamic_pvid(self):
        """Indicates whether a CNA can dynamically modify its PVID."""
        return not self._is_hmc

    @property
    def rmdev_job_available(self):
        """Indicates whether or not the Job API supports RMDev."""
        return not self._is_hmc

    @property
    def has_high_slot(self):
        """Does the API support UseNextAvailableHighSlotID?"""
        return not self._is_hmc

    @property
    def vea_as_ibmi_console(self):
        """Indicates whether the console type of IBMi VM is vea.

        IBMi depends on the trait to determine the console type. If the host is
        not managed by HMC, the console type of an IBMi VM deployed on the host
        shall be the slot number of its first virtual ethernet adapter.
        Otherwise, the Console type shall be "HMC".
        """
        return not self._is_hmc



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pypowervm.tests.test_utils import test_wrapper_abc as twrap
import pypowervm.wrappers.management_console as mc

_PUB_KEY = ('ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCuA/Av0jMYlG54YiaaaQXho8iO'
            'ZfY+WkBnuFfweESZOy824Ce9FvPqXsNL+nPAgKWG3TONwJldYgCgnBsFXUizkcne'
            '9Dt/T/zs2Bzl7b1YPrXyYS1hxKFrV/pYEERUiFa9ppR+M8mxdNYO0+ph356LO3mb'
            'xOM6nEZ1L6l6RUvbUwV9Zuw3Hpiz1lAV6d6EwMHJZ+WFlipJ2wxpM4QUKmb0V2UJ'
            'oHAb7tp3zipr3CCo0NtnpcD7wxsFhtz2ccRvNMbGhe1i9KikmBtQQDl1adMSbBL2'
            '+tGmyqHNq/H6d75bfXOUCl7NKtUq7VVGcXDOlTS1CDdLdmUn0l4z0AlyciQt wlp'
            '@9.0.0.0')
_AUTH_KEYS = ('ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCyl3+yXyoYAzvScxTiWqxe0P'
              'DwYvTHwLsIkgAY7s7n+8tUR7zA0dYWggl4aCfOAE2RMF0zKoFyRK8a9M/I1kVC'
              'YLb9y1rWp76jnxZpRBD/1DjjQ0qW5e1fbdrS52mJcFLL1+MzeoLT7+6GeMUcgN'
              'rmZQMUqSbwF+Rdxv56YTdx9u0EH1qaT/H0syp1Y8EHCaBVwdZcmNQLBFaYnVxH'
              'NHTQMYMTqokkyrZ9whSaK98OiYQO//5gnJzESOxOURYTzLKLz8WPkiONM6QgF+'
              'E5Zobt/REr3Tq8l1e1V/e2+7owFkMMte14I2sfK8QnZUrpJziXv3gwOpUP34gD'
              'ud6ceBlv wlp@9.0.0.0',
              'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQC3shE8yLGII+BPaPMIOdNgA6'
              'ZyDYKobCtXE6td8X9dgI0Sz08YCUQY9pOeWr/D63LwJaYsgqVspQaUEM5WH6s2'
              'eNKAERYayog6iCEaqApDDQETuf4XQ0JXo08izRPpMeRZwp3/RhNJVrxNheUp9n'
              'kHI3Mbx7jHvgwih48BTeqfj8L1Nnp4srhYDuzuN6NhUvbWLKJAjaQojRLSYEty'
              's5ASq7v+D+OEXqVBSRheKf5eWOdEF68sBYpOaS4qLycZjd5YGPUg0b+DfME2jr'
              '8kjbig1js8omgljSvKIwHIKfrfWPwKbWxtHaqWzTT+fUPygD7IDxPqsSEQIAjN'
              'PWmWQM+D wlp@9.0.0.0')


class TestMCEntryWrapper(twrap.TestWrapper):
    file = 'managementconsole.txt'
    wrapper_class_to_test = mc.ManagementConsole

    def test_mgmt_console(self):
        self.assertEqual(self.dwrap.name, "hmc7")
        self.assertEqual(self.dwrap.mtms.model, "f93")

        self.assertEqual(self.dwrap.mtms.machine_type, "Ve57")

        self.assertEqual(self.dwrap.mtms.serial, "2911559")

        con_inf = self.dwrap.network_interfaces.console_interface
        self.assertEqual('eth0', con_inf.name)
        self.assertEqual('9.1.2.3 fe80:0:0:0:5054:ff:fed8:a951',
                         con_inf.address)

        self.assertEqual(None, self.dwrap.ssh_public_key)
        self.assertEqual((), self.dwrap.ssh_authorized_keys)


class TestMCEntryWrapperSSH(twrap.TestWrapper):
    file = 'managementconsole_ssh.txt'
    wrapper_class_to_test = mc.ManagementConsole

    def test_mgmt_console(self):
        self.assertEqual(_PUB_KEY, self.dwrap.ssh_public_key)
        self.assertEqual(_AUTH_KEYS, self.dwrap.ssh_authorized_keys)

        # Now set the keys
        keys = ('key1', 'key2')
        self.dwrap.ssh_authorized_keys = keys
        self.assertEqual(keys, self.dwrap.ssh_authorized_keys)

        self.dwrap.ssh_authorized_keys = []
        self.assertEqual(tuple(), self.dwrap.ssh_authorized_keys)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import six
import testtools

import pypowervm.tests.test_fixtures as fx
from pypowervm.tests.test_utils import pvmhttp


@six.add_metaclass(abc.ABCMeta)
class TestWrapper(testtools.TestCase):
    """Superclass for wrapper test cases; provides loading of data files.

    A single subclass tests a single wrapper class on a single file.

    Usage:
    - Subclass this class.
    - Provide the name of the data file to load, e.g.
        file = 'ssp.txt'
    - Indicate the wrapper class to be tested, e.g.
        wrapper_class_to_test = clust.SSP
    - If your tests will make use of traits, you must provide
      mock_adapter_fx_args, resulting in AdapterFx being constructed with those
      args and used via useFixture.  Your tests may access the adapter via
      self.adpt and the fixture itself via self.adptfx.
    - No __init__ or setUp is necessary.
    - In your test cases, make use of the following variables:
        - self.resp: The raw Response object from
          load_pvm_resp().get_response().  May represent an entry or a feed.
        - self.dwrap: A single instance of the wrapper_class_to_test extracted
          from self.resp.  If self.resp was a feed, this is the first entry.
        - self.entries: The result of wrap(response) of the wrapper class.
          May be a single wrapper instance, in which case it's (nearly*)
          equivalent to self.dwrap, or a list of such wrappers.
          * Note that wrap(response) injects each entry's etag into the
            wrapper instance.
    """

    # Load the response file just once
    _pvmfile = None

    @abc.abstractproperty
    def file(self):
        """Data file name, relative to pypowervm/tests/wrappers/data/."""
        return None

    @abc.abstractproperty
    def wrapper_class_to_test(self):
        """Indicates the type (Wrapper subclass) produced by self.dwrap."""
        return None

    # Arguments to test_fixtures.AdapterFx(), used to create a mock adapter.
    # Must be represented as a dict.  For example:
    #     mock_adapter_fx_args = {}
    # or:
    #     mock_adapter_fx_args = dict(session=mock_session,
    #                                 traits=test_fixtures.LocalPVMTraits)
    mock_adapter_fx_args = None

    def setUp(self):
        super(TestWrapper, self).setUp()
        self.adptfx = None
        self.adpt = None
        adptfx_args = self.mock_adapter_fx_args or {}
        self.adptfx = self.useFixture(fx.AdapterFx(**adptfx_args))
        self.adpt = self.adptfx.adpt

        # Load the file just once...
        if self.__class__._pvmfile is None:
            self.__class__._pvmfile = pvmhttp.PVMFile(self.file)
        # ...but reconstruct the PVMResp every time
        self.resp = pvmhttp.PVMResp(pvmfile=self.__class__._pvmfile,
                                    adapter=self.adpt).get_response()
        # Some wrappers don't support etag.  Subclasses testing those wrappers
        # should not be using self.entries, so ignore.
        try:
            self.entries = self.wrapper_class_to_test.wrap(self.resp)
        except TypeError:
            pass
        if self.resp.feed:
            self.dwrap = self.wrapper_class_to_test.wrap(
                self.resp.feed.entries[0])
        else:
            self.dwrap = self.wrapper_class_to_test.wrap(self.resp.entry)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import mock
import requests.models as req_mod
import requests.structures as req_struct

import pypowervm.adapter as adp
import pypowervm.tests.lib as testlib
from pypowervm.tests.test_utils import pvmhttp
from pypowervm import traits
import pypowervm.wrappers.entry_wrapper as ewrap
import pypowervm.wrappers.network as net
import pypowervm.wrappers.storage as stor

_logon_response_text = testlib.file2b("logon_file.xml")
_feed_file = pvmhttp.load_pvm_resp(
    "fake_network_bridge.txt").response.body.encode('utf-8')
_entry_file = pvmhttp.load_pvm_resp(
    "fake_volume_group.txt").response.body.encode('utf-8')


class TestTraits(unittest.TestCase):

    @mock.patch('pypowervm.adapter.Session')
    def test_traits(self, mock_sess):
        # PVM MC, local auth
        mock_sess.mc_type = 'PVM'
        mock_sess.use_file_auth = True
        t = traits.APITraits(mock_sess)
        self.assertFalse(t.vnet_aware)
        self.assertFalse(t._is_hmc)
        self.assertTrue(t.local_api)
        self.assertFalse(t.has_lpar_profiles)
        self.assertTrue(t.dynamic_pvid)
        self.assertTrue(t.rmdev_job_available)
        self.assertTrue(t.has_high_slot)
        self.assertTrue(t.vea_as_ibmi_console)

        # PVM MC, remote auth
        mock_sess.mc_type = 'PVM'
        mock_sess.use_file_auth = False
        t = traits.APITraits(mock_sess)
        self.assertFalse(t.vnet_aware)
        self.assertFalse(t._is_hmc)
        self.assertFalse(t.local_api)
        self.assertFalse(t.has_lpar_profiles)
        self.assertTrue(t.dynamic_pvid)
        self.assertTrue(t.rmdev_job_available)
        self.assertTrue(t.has_high_slot)
        self.assertTrue(t.vea_as_ibmi_console)

        # HMC, remote auth
        mock_sess.mc_type = 'HMC'
        mock_sess.use_file_auth = False
        t = traits.APITraits(mock_sess)
        self.assertTrue(t.vnet_aware)
        self.assertTrue(t._is_hmc)
        self.assertFalse(t.local_api)
        self.assertTrue(t.has_lpar_profiles)
        self.assertFalse(t.dynamic_pvid)
        self.assertFalse(t.rmdev_job_available)
        self.assertFalse(t.has_high_slot)
        self.assertFalse(t.vea_as_ibmi_console)

    @mock.patch('requests.Session.request')
    def test_traits_into_wrappers(self, mock_request):
        # Note traits param is None, which reflects the real value of
        # self.traits during _logon's request.
        httpresp = req_mod.Response()
        httpresp._content = _logon_response_text
        httpresp.status_code = 200
        httpresp.headers = req_struct.CaseInsensitiveDict(
            {'X-MC-Type': 'PVM',
             'content-type':
                 'application/vnd.ibm.powervm.web+xml; type=LogonResponse'})
        mock_request.return_value = httpresp
        sess = adp.Session()
        self.assertEqual('PVM', sess.mc_type)
        self.assertIsNotNone(sess.traits)
        self.assertTrue(sess.traits.local_api)
        self.assertFalse(sess.traits._is_hmc)
        adapter = adp.Adapter(sess)
        self.assertEqual(sess.traits, adapter.traits)

        # Response => Feed => Entrys => EntryWrappers => sub-ElementWrappers
        httpresp._content = _feed_file
        resp = adapter.read('NetworkBridge')
        self.assertEqual(sess.traits, resp.adapter.traits)
        nblist = net.NetBridge.wrap(resp)
        for nb in nblist:
            self.assertIsInstance(nb, net.NetBridge)
            self.assertEqual(sess.traits, nb.traits)
        seas = nblist[0].seas
        for sea in seas:
            self.assertIsInstance(sea, net.SEA)
            self.assertEqual(sess.traits, sea.traits)
        trunk = seas[0].primary_adpt
        self.assertIsInstance(trunk, net.TrunkAdapter)
        self.assertEqual(sess.traits, trunk.traits)

        # Response => Entry => EntryWrapper => sub-EntryWrappers
        # => sub-sub-ElementWrapper
        httpresp._content = _entry_file
        resp = adapter.read('VolumeGroup', root_id='abc123')
        self.assertEqual(sess.traits, resp.adapter.traits)
        vgent = stor.VG.wrap(resp)
        self.assertIsInstance(vgent, stor.VG)
        self.assertEqual(sess.traits, vgent.traits)
        pvs = vgent.phys_vols
        for pvent in pvs:
            self.assertIsInstance(pvent, stor.PV)
            self.assertEqual(sess.traits, pvent.traits)

        # Building raw wrappers from scratch
        class MyEntryWrapper(ewrap.EntryWrapper):
            schema_type = 'SomeObject'

            @classmethod
            def bld(cls, adpt):
                return super(MyEntryWrapper, cls)._bld(adpt)

        mew = MyEntryWrapper.bld(adapter)
        self.assertIsInstance(mew, MyEntryWrapper)
        self.assertEqual(sess.traits, mew.traits)

        class MyElementWrapper(ewrap.ElementWrapper):
            schema_type = 'SomeObject'

            @classmethod
            def bld(cls, adpt):
                return super(MyElementWrapper, cls)._bld(adpt)

        mew = MyElementWrapper.bld(adapter)
        self.assertIsInstance(mew, MyElementWrapper)
        self.assertEqual(sess.traits, mew.traits)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test for the monitoring functions."""

import datetime

import mock
import testtools

from pypowervm import entities as pvm_e
from pypowervm.tasks.monitor import util as pvm_t_mon
from pypowervm.tests.tasks import util as tju
from pypowervm.tests import test_fixtures as fx
from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers import monitor as pvm_mon
from pypowervm.wrappers.pcm import lpar as pvm_mon_lpar
from pypowervm.wrappers.pcm import phyp as pvm_mon_phyp
from pypowervm.wrappers.pcm import vios as pvm_mon_vios


PHYP_DATA = 'phyp_pcm_data.txt'
VIOS_DATA = 'vios_pcm_data.txt'
LPAR_DATA = 'lpar_pcm_data.txt'
LTM_FEED = 'ltm_feed2.txt'


class TestMonitors(testtools.TestCase):

    def setUp(self):
        super(TestMonitors, self).setUp()

        self.adptfx = self.useFixture(fx.AdapterFx(traits=fx.RemoteHMCTraits))
        self.adpt = self.adptfx.adpt

    def test_query_ltm_feed(self):
        self.adpt.read_by_path.return_value = tju.load_file(LTM_FEED)
        feed = pvm_t_mon.query_ltm_feed(self.adpt, 'host_uuid')

        # Make sure the feed is correct.  Our sample data has 130 elements
        # in the feed.
        self.assertEqual(130, len(feed))

        # Make sure each element is a LTMMetric
        for mon in feed:
            self.assertIsInstance(mon, pvm_mon.LTMMetrics)

        self.assertEqual(1, self.adpt.read_by_path.call_count)

    def test_ensure_ltm_monitors(self):
        """Verifies that the LTM monitors can be turned on."""
        resp = tju.load_file('pcm_pref.txt')
        self.adpt.read_by_href.return_value = resp

        # Create a side effect that can validate the input to the update
        def validate_of_update(*kargs, **kwargs):
            element = kargs[0]
            etag = kargs[1]
            self.assertIsNotNone(element)
            self.assertEqual('-215935973', etag)

            # Wrap the element so we can validate it.
            pref = pvm_mon.PcmPref.wrap(pvm_e.Entry({'etag': etag},
                                                    element, self.adpt))

            self.assertFalse(pref.compute_ltm_enabled)
            self.assertTrue(pref.ltm_enabled)
            self.assertFalse(pref.stm_enabled)
            self.assertFalse(pref.aggregation_enabled)
            return element
        self.adpt.update.side_effect = validate_of_update

        # This will invoke the validate_of_update
        pvm_t_mon.ensure_ltm_monitors(self.adpt, 'host_uuid')

        # Make sure the update was in fact invoked though
        self.assertEqual(1, self.adpt.update.call_count)

    def test_ensure_ltm_monitors_non_default(self):
        """Verifies that the LTM monitors with different default inputs"""
        resp = tju.load_file('pcm_pref.txt')
        self.adpt.read_by_href.return_value = resp

        # Create a side effect that can validate the input to the update
        def validate_of_update(*kargs, **kwargs):
            element = kargs[0]
            etag = kargs[1]
            self.assertIsNotNone(element)

            # Wrap the element so we can validate it.
            pref = pvm_mon.PcmPref.wrap(pvm_e.Entry({'etag': etag},
                                                    element, self.adpt))

            self.assertTrue(pref.compute_ltm_enabled)
            self.assertTrue(pref.ltm_enabled)
            self.assertFalse(pref.stm_enabled)
            self.assertTrue(pref.aggregation_enabled)
            return element
        self.adpt.update.side_effect = validate_of_update

        # This will invoke the validate_of_update
        pvm_t_mon.ensure_ltm_monitors(self.adpt, 'host_uuid', compute_ltm=True,
                                      override_to_default=True)

        # Make sure the update was in fact invoked though
        self.assertEqual(1, self.adpt.update.call_count)

    def _load(self, file_name):
        """Loads a file."""
        return pvmhttp.PVMFile(file_name).body

    def test_parse_to_vm_metrics(self):
        """Verifies the parsing to LPAR metrics."""
        phyp_resp = self._load(PHYP_DATA)
        phyp_data = pvm_mon_phyp.PhypInfo(phyp_resp)

        vios_resp = self._load(VIOS_DATA)
        vios_data = pvm_mon_vios.ViosInfo(vios_resp)

        lpar_resp = self._load(LPAR_DATA)
        lpar_data = pvm_mon_lpar.LparInfo(lpar_resp)

        metrics = pvm_t_mon.vm_metrics(phyp_data, [vios_data], lpar_data)
        self.assertIsNotNone(metrics)

        # In the test data, there are 5 LPARs total.
        self.assertEqual(5, len(metrics.keys()))

        # Validate a metric with live data
        good_vm = '42AD4FD4-DC64-4935-9E29-9B7C6F35AFCC'
        metric = metrics.get(good_vm)
        self.assertIsNotNone(metric)

        self.assertIsNotNone(metric.network)
        self.assertIsNotNone(metric.storage)
        self.assertIsNotNone(metric.processor)
        self.assertIsNotNone(metric.memory)

        # Memory validation
        self.assertEqual(20480, metric.memory.logical_mem)
        self.assertEqual(20480, metric.memory.backed_physical_mem)
        self.assertEqual(80, metric.memory.pct_real_mem_avbl)
        self.assertEqual(1024, metric.memory.total_pg_count)
        self.assertEqual(512, metric.memory.free_pg_count)
        self.assertEqual(1048576, metric.memory.real_mem_size_bytes)
        self.assertEqual(61, metric.memory.pct_real_mem_free)
        self.assertEqual(25, metric.memory.vm_pg_out_rate)

        # Processor validation
        self.assertEqual(0, metric.processor.pool_id)
        self.assertEqual('uncap', metric.processor.mode)
        self.assertEqual(4, metric.processor.virt_procs)
        self.assertEqual(.4, metric.processor.proc_units)

        # Network validation
        self.assertEqual(1, len(metric.network.cnas))
        cna = metric.network.cnas[0]
        self.assertEqual(2227, cna.vlan_id)
        self.assertEqual(0, cna.vswitch_id)
        self.assertEqual('U8247.22L.2125D4A-V2-C2', cna.physical_location)
        self.assertEqual(10, cna.received_packets)
        self.assertEqual(100, cna.sent_packets)
        self.assertEqual(5, cna.dropped_packets)
        self.assertEqual(100, cna.sent_bytes)
        self.assertEqual(10000, cna.received_bytes)

        # Storage validation
        self.assertEqual(1, len(metric.storage.virt_adpts))
        self.assertEqual(0, len(metric.storage.vfc_adpts))
        vadpt = metric.storage.virt_adpts[0]
        self.assertEqual('virtual', vadpt.type)
        self.assertEqual('vhost0', vadpt.name)
        self.assertEqual('U8247.22L.2125D4A-V1-C1000', vadpt.physical_location)
        self.assertEqual(1074, vadpt.num_reads)
        self.assertEqual(1075, vadpt.num_writes)
        self.assertEqual(549888, vadpt.read_bytes)
        self.assertEqual(550400, vadpt.write_bytes)

        # Validate a metric for a system that was powered off.
        bad_vm = '3B0237F9-26F1-41C7-BE57-A08C9452AD9D'
        metric = metrics.get(bad_vm)
        self.assertIsNotNone(metric)

        self.assertIsNotNone(metric.processor)
        self.assertIsNotNone(metric.memory)
        # For powered off VM, OS specific memory metrics are None
        self.assertIsNone(metric.memory.pct_real_mem_avbl)
        self.assertIsNone(metric.memory.total_pg_count)
        self.assertIsNone(metric.memory.free_pg_count)
        self.assertIsNone(metric.memory.active_pg_count)
        self.assertIsNone(metric.memory.real_mem_size_bytes)
        # For powered off VM, the free memory is 100 percent.
        self.assertEqual(100, metric.memory.pct_real_mem_free)
        # For powered off VM, the page in/out rate is 0.
        self.assertEqual(0, metric.memory.vm_pg_out_rate)
        self.assertIsNone(metric.storage)
        self.assertIsNone(metric.network)

        # Take a VM which has entry in phyp data but not in PCM Lpar data.
        # Assert that it has been correctly parsed and memory metrics
        # are set to default values.
        vm_in_phyp_not_in_lpar_pcm = '66A2E886-D05D-42F4-87E0-C3BA02CF7C7E'
        metric = metrics.get(vm_in_phyp_not_in_lpar_pcm)
        self.assertIsNotNone(metric)
        self.assertIsNotNone(metric.processor)
        self.assertIsNotNone(metric.memory)
        self.assertEqual(.2, metric.processor.proc_units)
        self.assertEqual(0, metric.memory.pct_real_mem_free)
        self.assertEqual(-1, metric.memory.vm_pg_in_rate)

    def test_vm_metrics_no_phyp_data(self):
        self.assertEqual({}, pvm_t_mon.vm_metrics(None, [], None))

    @mock.patch('pypowervm.tasks.monitor.util.query_ltm_feed')
    def test_latest_stats(self, mock_ltm_feed):
        # Set up the return data.
        mock_phyp_metric = mock.MagicMock()
        mock_phyp_metric.category = 'phyp'
        mock_phyp_metric.updated_datetime = 1
        mock_phyp_metric.link = 'bad'

        mock_phyp2_metric = mock.MagicMock()
        mock_phyp2_metric.category = 'phyp'
        mock_phyp2_metric.updated_datetime = 2
        mock_phyp2_metric.link = 'phyp'

        mock_vio1_metric = mock.MagicMock()
        mock_vio1_metric.category = 'vios_1'
        mock_vio1_metric.updated_datetime = 1
        mock_vio1_metric.link = 'bad'

        mock_vio2_metric = mock.MagicMock()
        mock_vio2_metric.category = 'vios_1'
        mock_vio2_metric.updated_datetime = 2
        mock_vio2_metric.link = 'vio'

        mock_vio3_metric = mock.MagicMock()
        mock_vio3_metric.category = 'vios_3'
        mock_vio3_metric.updated_datetime = 2
        mock_vio3_metric.link = 'vio'

        mock_lpar1_metric = mock.MagicMock()
        mock_lpar1_metric.category = 'lpar'
        mock_lpar1_metric.updated_datetime = 2
        mock_lpar1_metric.link = 'lpar'

        # Reset as this was invoked once up front.
        mock_ltm_feed.reset_mock()
        mock_ltm_feed.return_value = [mock_phyp_metric, mock_phyp2_metric,
                                      mock_vio1_metric, mock_vio2_metric,
                                      mock_vio3_metric, mock_lpar1_metric]

        # Data for the responses.
        phyp_resp = self._load(PHYP_DATA)
        vios_resp = self._load(VIOS_DATA)

        def validate_read(link, xag=None):
            resp = mock.MagicMock()
            if link == 'phyp':
                resp.body = phyp_resp
                return resp
            elif link == 'vio':
                resp.body = vios_resp
                return resp
            elif link == 'lpar':
                resp.body = self._load(LPAR_DATA)
                return resp
            else:
                self.fail()

        self.adpt.read_by_href.side_effect = validate_read

        resp_date, resp_phyp, resp_vioses, resp_lpars = (
            pvm_t_mon.latest_stats(self.adpt, mock.Mock()))
        self.assertIsNotNone(resp_phyp)
        self.assertIsInstance(resp_phyp, pvm_mon_phyp.PhypInfo)
        self.assertEqual(2, len(resp_vioses))
        self.assertIsInstance(resp_vioses[0], pvm_mon_vios.ViosInfo)
        self.assertIsInstance(resp_vioses[1], pvm_mon_vios.ViosInfo)
        self.assertEqual(6, len(resp_lpars.lpars_util))
        self.assertIsInstance(resp_lpars, pvm_mon_lpar.LparInfo)

        # Invoke again, but set to ignore vioses
        resp_date, resp_phyp, resp_vioses, resp_lpars = (
            pvm_t_mon.latest_stats(self.adpt, mock.Mock(), include_vio=False))
        self.assertIsNotNone(resp_phyp)
        self.assertIsInstance(resp_phyp, pvm_mon_phyp.PhypInfo)
        self.assertEqual(0, len(resp_vioses))

    @mock.patch('pypowervm.tasks.monitor.util.vm_metrics')
    @mock.patch('pypowervm.tasks.monitor.util.query_ltm_feed')
    def test_latest_stats_no_data(self, mock_ltm_feed, mock_vm_metrics):
        # Set up the return data.
        mock_vio3_metric = mock.MagicMock()
        mock_vio3_metric.category = 'vios_3'
        mock_vio3_metric.updated_datetime = 2

        # Reset as this was invoked once up front.
        mock_ltm_feed.reset_mock()
        mock_ltm_feed.return_value = [mock_vio3_metric]

        # Call the system.
        resp_date, resp_phyp, resp_vios, resp_lpars = (
            pvm_t_mon.latest_stats(mock.Mock(), mock.Mock()))
        self.assertIsNotNone(resp_date)
        self.assertIsNone(resp_phyp)
        self.assertIsNone(resp_vios)
        self.assertIsNone(resp_lpars)


class TestMetricsCache(testtools.TestCase):
    """Validates the LparMetricCache."""

    def setUp(self):
        super(TestMetricsCache, self).setUp()

        self.adptfx = self.useFixture(fx.AdapterFx(traits=fx.RemoteHMCTraits))
        self.adpt = self.adptfx.adpt

    @mock.patch('pypowervm.tasks.monitor.util.vm_metrics')
    @mock.patch('pypowervm.tasks.monitor.util.latest_stats')
    @mock.patch('pypowervm.tasks.monitor.util.ensure_ltm_monitors')
    def test_refresh(self, mock_ensure_monitor, mock_stats,
                     mock_vm_metrics):
        ret1 = None
        ret2 = {'lpar_uuid': 2}
        ret3 = {'lpar_uuid': 3}

        date_ret1 = datetime.datetime.now()
        date_ret2 = date_ret1 + datetime.timedelta(milliseconds=250)
        date_ret3 = date_ret2 + datetime.timedelta(milliseconds=250)

        mock_stats.side_effect = [
            (date_ret1, mock.Mock(), mock.Mock(), mock.Mock()),
            (date_ret2, mock.Mock(), mock.Mock(), mock.Mock()),
            (date_ret3, mock.Mock(), mock.Mock(), mock.Mock())]
        mock_vm_metrics.side_effect = [ret1, ret2, ret3]

        # Creation invokes the refresh once automatically.
        metric_cache = pvm_t_mon.LparMetricCache(self.adpt, 'host_uuid',
                                                 refresh_delta=.25)

        # Make sure the current and prev are none.
        self.assertEqual(date_ret1, metric_cache.cur_date)
        self.assertIsNone(metric_cache.cur_metric)
        self.assertIsNone(metric_cache.prev_date)
        self.assertIsNone(metric_cache.prev_metric)

        # The current metric should detect that it hasn't been enough time
        # and pass us none.
        cur_date, cur_metric = metric_cache.get_latest_metric('lpar_uuid')
        self.assertEqual(date_ret1, cur_date)
        self.assertIsNone(cur_metric)
        prev_date, prev_metric = metric_cache.get_previous_metric('lpar_uuid')
        self.assertIsNone(prev_date)
        self.assertIsNone(prev_metric)

        # Force the update by stating we're older than we are.
        pre_date = metric_cache.cur_date - datetime.timedelta(milliseconds=250)
        metric_cache.cur_date = pre_date

        # Verify that we've incremented
        cur_date, cur_metric = metric_cache.get_latest_metric('lpar_uuid')
        self.assertEqual(date_ret2, cur_date)
        self.assertEqual(2, cur_metric)
        prev_date, prev_metric = metric_cache.get_previous_metric('lpar_uuid')
        self.assertEqual(pre_date, prev_date)
        self.assertIsNone(prev_metric)

        # Verify that if we set the date to now, we don't increment
        metric_cache.cur_date = datetime.datetime.now()
        cur_date, cur_metric = metric_cache.get_latest_metric('lpar_uuid')
        self.assertEqual(2, cur_metric)
        prev_date, prev_metric = metric_cache.get_previous_metric('lpar_uuid')
        self.assertEqual(pre_date, prev_date)
        self.assertIsNone(prev_metric)

        # Delay one more time.  Make sure the previous values are now set.
        pre_date = metric_cache.cur_date - datetime.timedelta(milliseconds=250)
        metric_cache.cur_date = pre_date

        cur_date, cur_metric = metric_cache.get_latest_metric('lpar_uuid')
        self.assertEqual(date_ret3, cur_date)
        self.assertEqual(3, cur_metric)
        prev_date, prev_metric = metric_cache.get_previous_metric('lpar_uuid')
        self.assertEqual(pre_date, prev_date)
        self.assertEqual(2, prev_metric)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import unittest

import mock

import pypowervm.adapter as adp
import pypowervm.exceptions as pvmex
from pypowervm.helpers import vios_busy
from pypowervm.tests.test_utils import pvmhttp

HTTPRESP_FILE = "fake_httperror.txt"
HTTPRESP_SA_FILE = "fake_httperror_service_unavail.txt"


class TestVIOSBusyHelper(unittest.TestCase):

    def setUp(self):
        super(TestVIOSBusyHelper, self).setUp()

        self.http_error = pvmhttp.load_pvm_resp(HTTPRESP_FILE)
        self.http_error_sa = pvmhttp.load_pvm_resp(HTTPRESP_SA_FILE)

    @mock.patch('pypowervm.adapter.Session')
    @mock.patch('pypowervm.helpers.vios_busy.SLEEP')
    def test_vios_busy_helper(self, mock_sleep, mock_sess):
        # Try with 1 retries
        hlp = functools.partial(vios_busy.vios_busy_retry_helper,
                                max_retries=1)
        error = pvmex.Error('yo', response=self.http_error.response)
        mock_sess.request.side_effect = error
        adpt = adp.Adapter(mock_sess, helpers=hlp)
        self.assertRaises(
            pvmex.Error, adpt._request, 'method', 'path', body='the body')
        # Test that the request method was called twice and sleep was called
        self.assertEqual(mock_sess.request.call_count, 2)
        mock_sleep.assert_called_once_with(5 * 1)

        # Test with more retries and sleep values
        retries = 10
        hlp = functools.partial(vios_busy.vios_busy_retry_helper,
                                max_retries=retries, delay=15)
        mock_sess.reset_mock()
        self.assertRaises(pvmex.Error, adpt._request, 'method', 'path',
                          body='the body', helpers=hlp)
        # Should have tried 'retries' times plus the initial one
        self.assertEqual(mock_sess.request.call_count, retries+1)

        # Test with None response
        mock_sess.reset_mock()
        error = pvmex.Error('yo', response=None)
        mock_sess.request.side_effect = error
        hlp = functools.partial(vios_busy.vios_busy_retry_helper,
                                max_retries=1, delay=15)
        self.assertRaises(pvmex.Error, adpt._request, 'method', 'path',
                          body='the body', helpers=hlp)
        # There should be no retries since the response was None
        self.assertEqual(mock_sess.request.call_count, 1)

        # Test with a Service Unavailable exception
        mock_sess.reset_mock()
        hlp = functools.partial(vios_busy.vios_busy_retry_helper,
                                max_retries=1)
        error = pvmex.Error('yo', response=self.http_error_sa.response)
        mock_sess.request.side_effect = error
        adpt = adp.Adapter(mock_sess, helpers=hlp)
        self.assertRaises(
            pvmex.Error, adpt._request, 'method', 'path', body='the body')
        self.assertEqual(mock_sess.request.call_count, 2)
